# coding:utf-8
from typing import Optional as Opt
import json
from logging import Logger, Manager, addLevelName, getLevelName, root
from collections import Counter
from .defaults import PROGRESS, ERROR, PROGRESS_LEVEL, METALOG, METALOG_LEVEL
from .timing import TimedTaskContext
from .helpers import *

logger_method_names = ("debug", "info", "warning", "error", "critical")


class CountingLogger(Logger):
    """
    This logger subclass allows validation of log file parses by counting the number of
    times that it has logged at each level, how many stack traces it has logged, and how
    many multiline messages it has logged (which are potentially ambiguous to parse)
    """

    addLevelName(METALOG, METALOG_LEVEL)
    manager = Manager(root)

    def __init__(self, *args, **kwargs):
        self.multiline = 0
        self.stacktraces = 0
        self.total = 0
        self.levelcounts = Counter()
        super().__init__(*args, **kwargs)

    def _log(
        self, level, msg, args, exc_info=None, extra=None, stack_info=False, stats=False
    ):
        levelname = getLevelName(level)

        if not stats:
            super()._log(
                level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info
            )
            if exc_info:
                self.stacktraces += 1
            if "\n" in msg:
                self.multiline += 1

        if level >= self.level:
            self.levelcounts.update([levelname])

        if stats:
            msg = json.dumps(self.stats, indent=None)
            if "\n" in msg:
                self.multiline += 1
                stats_ = json.loads(msg)
                stats_["multiline"] += 1
                msg = json.dumps(self.stats, indent=None)
            super()._log(
                level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info
            )

        self.total += 1

    @property
    def stats(self):
        return dict(
            stacktraces=self.stacktraces,
            multiline=self.multiline,
            levelcounts=self.levelcounts,
        )

    def _report_stats(self):
        self._log(METALOG, None, None, stats=True)

    def __del__(self):
        # log the final message count stats on program exit
        self._report_stats()


class ProgressLogger(CountingLogger):
    """
    This logger subclass allows for simple and consistent logging of progress on user-defined
    jobs:
    >>> logger = ProgressLogger("fooLog", job_level="INFO")
    >>> with logger.task("count_to_10", time_units='s') as task:
    >>>     for i in range(10):
    >>>         task.report_progress(1)
    >>>         # verbose output happens here
    """

    addLevelName(PROGRESS, PROGRESS_LEVEL)
    manager = Manager(root)

    def __init__(self, *args, job_level=PROGRESS_LEVEL, **kwargs):
        self.job_level = validate_log_level(job_level)
        super().__init__(*args, **kwargs)

    def task(
        self,
        job_name: str,
        total_tasks: Opt[int] = None,
        task_units: Opt[str] = None,
        level=PROGRESS,
        error_level=ERROR,
        time_units: str = "s",
    ):
        level = level or self.job_level
        return TimedTaskContext(
            job_name,
            total_tasks=total_tasks,
            task_units=task_units,
            time_units=time_units,
            logger_or_print_func=self,
            level=level,
            error_level=error_level,
        )


for cls in (CountingLogger, ProgressLogger):
    cls.manager.loggerClass = cls



# coding:utf-8
from typing import IO, Any, Mapping, Sequence, Union, Optional as Opt
import os
import argparse
from enum import Enum
from pathlib import Path
from functools import partial
from logging import getLogger
from collections import ChainMap
import yaml
import toml
import ujson as json
from bourbaki.introspection.prettyprint import has_identifier_keys
from ..namespace import namespace_recursive
from ..paths import get_file, ensure_dir, path_with_ext
from .python import load_python, dump_python, MAX_PY_WIDTH
from .ini import load_ini, dump_ini
from .exceptions import ConfigNotSerializable

NoneType = type(None)
logger = getLogger(__name__)

# config formatting constants

LEGAL_CONFIG_EXTENSIONS = {".yml", ".yaml", ".json", ".toml", ".py", ".ini"}
EMPTY_CONFIG_VALUE = "________"
# Force indentation for small collections for readability, don't sort keys to preserve method def order in CLIs
JSON_DUMP_KWARGS = dict(indent=2, sort_keys=False)
YAML_DUMP_KWARGS = dict(
    default_flow_style=False, width=MAX_PY_WIDTH, sort_keys=False, indent=2
)

loaders = {
    ".yml": yaml.safe_load,
    ".yaml": yaml.safe_load,
    ".json": json.load,
    ".toml": toml.load,
    ".py": load_python,
    ".ini": load_ini,
}
loader_kw = {}

dumpers = {
    ".yml": yaml.safe_dump,
    ".yaml": yaml.safe_dump,
    ".json": json.dump,
    ".toml": toml.dump,
    ".py": dump_python,
    ".ini": dump_ini,
}
dumper_kw = {
    ".json": JSON_DUMP_KWARGS,
    ".yml": YAML_DUMP_KWARGS,
    ".yaml": YAML_DUMP_KWARGS,
}


class ConfigFormat(Enum):
    yaml = yml = ".yml"
    toml = ".toml"
    json = ".json"
    ini = ".ini"
    py = ".py"


class UnknownConfigExtension(ValueError):
    def __init__(self, ext):
        super().__init__(ext)
        self.ext = ext

    def __str__(self):
        return "Unknown config extension: {}; legal choices are {}".format(
            repr(self.ext), tuple(sorted(LEGAL_CONFIG_EXTENSIONS))
        )


class UnknownConfigLoadExtension(UnknownConfigExtension):
    def __str__(self):
        return super().__str__() + "; cannot load"


class UnknownConfigDumpExtension(UnknownConfigExtension):
    def __str__(self):
        return super().__str__() + "; cannot dump"


def normalize_ext(ext):
    return "." + ext.lstrip(".")


def _register_config_io(exts, default_kw, func_registry, kw_registry):
    exts = list(map(normalize_ext, exts))

    def dec(f):
        for ext in exts:
            func_registry[ext] = f
            if default_kw:
                kw_registry[ext] = default_kw
            LEGAL_CONFIG_EXTENSIONS.add(ext)
        return f

    return dec


def register_load_config(*exts: str, **default_kw):
    """
    examples:

    @register_load_config(".jsonl", ".jsonlines", linebreak='\n')
    def load_json_lines(file_handle, linebreak):
        return list(map(json.loads, map(str.strip, file_handle.read().split(linebreak))))

    OR

    register_load_config(".yaml")(yaml.safe_load)
    """
    return _register_config_io(exts, default_kw, loaders, loader_kw)


def register_dump_config(*exts: str, **default_kw):
    """
    examples:

    @register_dump_config(".jsonl", ".jsonlines")
    def dump_json_lines(iterable, file_handle):
        for obj in iterable:
            print(json.dumps(obj, file=file_handle))

    OR

    register_dump_config(".yaml", indent=4)(yaml.safe_dump)
    """
    return _register_config_io(exts, default_kw, dumpers, dumper_kw)


def allow_unsafe_yaml():
    global loaders, dumpers
    for ext in (".yml", ".yaml"):
        loaders[ext] = yaml.load
    for ext in (".yml", ".yaml"):
        dumpers[ext] = yaml.dump


def require_safe_yaml():
    global loaders, dumpers
    for ext in (".yml", ".yaml"):
        loaders[ext] = yaml.safe_load
    for ext in (".yml", ".yaml"):
        dumpers[ext] = yaml.safe_dump


def _config_io(
    load: bool, obj: Any, file: IO, ext: str, kw: Opt[Mapping[str, Any]] = None
):
    if load:
        func_registry, kw_registry, exc_type, args = (
            loaders,
            loader_kw,
            UnknownConfigLoadExtension,
            (file,),
        )
    else:
        func_registry, kw_registry, exc_type, args = (
            dumpers,
            dumper_kw,
            UnknownConfigDumpExtension,
            (obj, file),
        )

    try:
        io_func = func_registry[ext]
    except KeyError:
        raise exc_type(ext)
    else:
        defaults = kw_registry.get(ext)
        kws = [d for d in (kw, defaults) if d]
        if not kws:
            result = io_func(*args)
        else:
            result = io_func(*args, **ChainMap(*kws))

    return result


def _load_config(file: IO, ext: str, kw: Opt[Mapping[str, Any]] = None):
    return _config_io(True, None, file, ext, kw)


def _dump_config(obj: Any, file: IO, ext: str, kw: Opt[Mapping[str, Any]] = None):
    return _config_io(False, obj, file, ext, kw)


def _load_config_dir(config_file, ext=None):
    load = partial(load_config, namespace=False, ext=ext, disambiguate=False)
    filenames = os.listdir(config_file)
    stripped_names = [os.path.splitext(p)[0] for p in filenames]
    qualified_names = [os.path.join(config_file, p) for p in filenames]
    return dict(zip(stripped_names, map(load, qualified_names)))


def load_config(
    config_file: Union[str, Path, IO],
    ext: Opt[str] = None,
    disambiguate: bool = False,
    namespace: bool = False,
    **load_kw
):
    # try to get a name for the file to dispatch on
    if isinstance(config_file, (Path, str)):
        if os.path.isdir(config_file) and disambiguate:
            # load config for each file in the dir
            conf = _load_config_dir(config_file, ext=ext)
            if namespace:
                conf = namespace_recursive(conf)
            return conf
        else:
            # try to load later for a single path
            filename = str(config_file)
    else:
        # try for an open file handle with name attribute (not all have it)
        try:
            filename = config_file.name
        except AttributeError:
            filename = None

    # Then we haven't parsed from a directory; continue with a file
    # if passed, ext determines the serialization protocol used, otherwise it is inferred from config_file
    if ext is None:
        if filename is None:
            raise ValueError(
                "To load config from filename {}, you must pass `ext` to specify the "
                "serialization protocol".format(config_file)
            )

        filename, ext = path_with_ext(filename, ext, disambiguate=disambiguate)
    else:
        filename = config_file if filename is None else filename

    file, close = get_file(filename, "r")

    if close:
        with file:
            conf = _load_config(file, ext, load_kw)
    else:
        conf = _load_config(file, ext, load_kw)

    if namespace:
        conf = namespace_recursive(conf)

    return conf


def dump_config(
    conf: Union[Mapping, Sequence, argparse.Namespace],
    config_file: Union[str, Path, IO],
    ext: Opt[Union[ConfigFormat, str]] = None,
    disambiguate: bool = False,
    as_dir: bool = False,
    allow_dir: bool = False,
    **dump_kw
):
    if not isinstance(conf, (Mapping, Sequence)):
        raise ConfigNotSerializable(
            "conf must be a Mapping or Sequence type; got {}".format(type(conf))
        )

    if isinstance(ext, ConfigFormat):
        ext = ext.value

    if as_dir:
        if ext is None:
            raise ValueError("must provide an extension when as_dir=True")
        if not isinstance(config_file, (Path, str)):
            raise TypeError(
                "cannot determine a directory from config_file arg of type {}; pass a str or pathlib.Path".format(
                    type(config_file)
                )
            )
        ensure_dir(config_file)
        allow_dir = True
    elif isinstance(config_file, (Path, str)):
        config_file, ext = path_with_ext(config_file, ext, disambiguate)
    elif ext is None:
        try:
            fname = config_file.name
        except AttributeError:
            pass
        else:
            ext = os.path.splitext(fname)[-1]

        if not ext:
            raise ValueError(
                "Could not infer config extension from file {}".format(config_file)
            )

    # file is an open file handle if not a dir, else a Path
    file, close = get_file(config_file, "w", allow_dir=allow_dir)

    if isinstance(file, Path):
        # directory
        if not isinstance(conf, Mapping) or not has_identifier_keys(conf):
            raise ConfigNotSerializable(
                "Can only dump config of type Mapping[str, Any] with identifier keys to a "
                "directory; got {}".format(type(conf))
            )
        for name, subconf in conf.items():
            dump_config(
                subconf, file / name, ext=ext, as_dir=False, allow_dir=True, **dump_kw
            )
    elif close:
        # file
        with file:
            _dump_config(conf, file, ext, dump_kw)
    else:
        # file
        _dump_config(conf, file, ext, dump_kw)



# coding:utf-8
from typing import Union, Optional, TextIO, BinaryIO, IO
from argparse import FileType, ZERO_OR_MORE, ONE_OR_MORE, OPTIONAL
import io
import typing
import types
import datetime
import decimal
import encodings
import fractions
import ipaddress
import numbers
import pathlib
import sys
import uuid
from urllib.parse import ParseResult as URL
from functools import lru_cache, singledispatch
from inspect import Parameter
from multipledispatch import Dispatcher
from bourbaki.introspection.types import (
    deconstruct_generic,
    is_named_tuple_class,
    get_constructor_for,
    PseudoGenericMeta,
)
from bourbaki.introspection.typechecking import isinstance_generic
from bourbaki.introspection.docstrings import CallableDocs, ParamDocs, ParamDoc
from bourbaki.introspection.imports import import_object
from bourbaki.introspection.callables import function_classpath, UnStarred
from bourbaki.introspection.classes import classpath, parameterized_classpath
from bourbaki.introspection.generic_dispatch_helpers import PicklableWithType
from .exceptions import TypedInputError, TypedOutputError

Empty = Parameter.empty
Doc = Union[CallableDocs, ParamDocs, ParamDoc]

READ_MODES = {"r", "rb", "r+", "rb+", "a+", "ab+", "wb+"}
WRITE_MODES = {"w", "wb", "w+", "wb+", "a", "ab", "a+", "ab+", "rb+", "x", "xb"}
FILE_MODES = READ_MODES.union(WRITE_MODES)

NARGS_OPTIONS = (ZERO_OR_MORE, ONE_OR_MORE, OPTIONAL, None)
CLI_PREFIX_CHAR = "-"
KEY_VAL_JOIN_CHAR = "="


def unq(it):
    cache = set()
    for i in it:
        if i not in cache:
            cache.add(i)
            yield i


class _FileHandleConstructor(PseudoGenericMeta):
    @lru_cache(None)
    def __getitem__(cls, mode_enc) -> type:
        if isinstance(mode_enc, tuple):
            mode, encoding = mode_enc
        else:
            mode, encoding = mode_enc, None

        if cls.mode is not None or cls.encoding is not None:
            raise TypeError(
                "Can't subscript File more than once; tried to subscript {} with {}".format(
                    repr(cls), mode_enc
                )
            )
        mode, encoding, is_binary, base = _file_args(mode, encoding)
        new_cls = BinaryFile if is_binary else TextFile
        mcs = type(cls)
        return type.__new__(
            mcs,
            new_cls.__name__,
            (new_cls, base),
            dict(__args__=(mode, encoding), __origin__=File),
        )

    def __repr__(cls):
        tname = cls.__name__
        if cls.encoding:
            return "{}[{}, {}]".format(tname, repr(cls.mode), repr(cls.encoding))
        if cls.mode:
            return "{}[{}]".format(tname, repr(cls.mode))
        else:
            return tname

    @property
    def readable(cls):
        return cls.mode in READ_MODES

    @property
    def writable(cls):
        return cls.mode in WRITE_MODES

    @property
    def binary(cls):
        return "b" in cls.mode

    @property
    def mode(cls):
        if not cls.__args__:
            return None
        return cls.__args__[0]

    @property
    def encoding(cls):
        if not cls.__args__:
            return None
        return cls.__args__[1]

    def __instancecheck__(cls, instance):
        return isinstance(instance, cls.__bases__)


class File(metaclass=_FileHandleConstructor):
    encoding = None
    mode = None

    def __new__(cls, path) -> IO:
        # argparse.FileType has the nice feature of treating '-' as stdin/stdout and raising nice errors
        return FileType(cls.mode, encoding=cls.encoding)(path)


class TextFile(File):
    def __new__(cls, path) -> TextIO:
        return super().__new__(cls, path)


class BinaryFile(File):
    def __new__(cls, path) -> BinaryIO:
        return super().__new__(cls, path)


class PositionalMetavarFormatter:
    """Hack to deal with the fact that argparse doesn't allow tuples for positional arg metavars 
    (in contrast to the behavior for options)"""

    def __init__(self, *metavar: str, name: str):
        self.metavar = metavar
        self.name = name
        self._iter = iter(metavar * 2)

    def copy(self):
        return self.__class__(*self.metavar, name=self.name)

    def __getitem__(self, item):
        return self.metavar[item]

    def __str__(self):
        try:
            metavar = str(next(self._iter))
        except StopIteration:
            # for the help section
            s = self.name
        else:
            # for the usage line
            if self.name is None:
                s = metavar
            else:
                s = "{}-{}".format(self.name, metavar)

        return str(s)

    def __len__(self):
        return len(self.metavar)  # max(map(len, self.metavar))


def validate_nargs(nargs):
    if isinstance(nargs, int):
        if nargs > 0:
            return nargs
        raise ValueError("if an int, nargs must be positive; got {}".format(nargs))
    if nargs not in NARGS_OPTIONS:
        raise ValueError(
            "nargs must be a positive int or one of {}; got {}".format(
                NARGS_OPTIONS, nargs
            )
        )
    return nargs


def to_cmd_line_name(s: str, negative_flag=False):
    name = s.replace("_", "-").rstrip("-")
    if negative_flag:
        name = "no-" + name
    return name


def cmd_line_arg_names(name, positional=False, prefix_char=None, negative_flag=False):
    if not positional:
        if prefix_char is None:
            prefix_char = CLI_PREFIX_CHAR
        names = (
            (prefix_char * min(len(name), 2)) + to_cmd_line_name(name, negative_flag),
        )
    else:
        names = (name,)
    return names


def get_dest_name(args, prefix_chars):
    cs = prefix_chars
    try:
        dest = next(
            argname.lstrip(cs) for argname in args if len(argname.lstrip(cs)) > 1
        )
    except StopIteration:
        dest = args[0]
    return dest


@singledispatch
def to_str_cli_repr(repr_, n: Optional[int] = None):
    if n == ONE_OR_MORE:
        "{r} [{r} ...]".format(r=repr_)
    elif n == ZERO_OR_MORE:
        return "[{r} [{r} ...]]".format(r=repr_)
    elif n == OPTIONAL:
        return "[{}]".format(repr_)
    return repr_


@to_str_cli_repr.register(tuple)
@to_str_cli_repr.register(list)
def to_str_cli_repr_tuple(repr_, n):
    return " ".join(map(to_str_cli_repr, repr_))


def normalize_encoding(enc):
    if enc is None:
        return sys.getdefaultencoding()
    enc_ = encodings.search_function(enc)
    if not enc_:
        raise ValueError(
            "{} is not a valid text encoding; see encodings.aliases.aliases for the set of legal "
            "values".format(enc)
        )
    return enc_.name


def cached_property(method):
    """compute the getter once and store the result in the instance dict under the same name,
    so it isn't computed again"""
    attr = method.__name__

    def newmethod(self):
        value = method(self)
        self.__dict__[attr] = value
        return value

    return property(newmethod)


def normalize_file_mode(mode):
    if mode not in FILE_MODES:
        raise ValueError(
            "{} is not a valid file mode; choose one of {}".format(
                mode, tuple(FILE_MODES)
            )
        )
    return mode


def is_binary_mode(mode):
    return "b" in mode and mode in FILE_MODES


def is_write_mode(mode):
    return mode in WRITE_MODES


def is_read_mode(mode):
    return mode in READ_MODES


def _file_args(mode, encoding):
    mode = normalize_file_mode(mode)

    if is_binary_mode(mode):
        is_binary = True

        if encoding is not None:
            raise ValueError(
                "binary mode {} can't be specified with an encoding; got encoding={}".format(
                    repr(mode), repr(encoding)
                )
            )

        if is_read_mode(mode) and is_write_mode(mode):
            base = io.BufferedRandom
        elif is_read_mode(mode):
            base = io.BufferedReader
        else:
            base = io.BufferedWriter
    else:
        is_binary = False
        base = io.TextIOWrapper
        encoding = normalize_encoding(encoding)

    return mode, encoding, is_binary, base


def singleton(cls):
    return cls()


def identity(x):
    return x


def name_of(x):
    return getattr(x, "__name__", getattr(type(x), "__name__", str(x)))


@singledispatch
def type_spec(type_):
    if type_.__module__ == "builtins":
        path = type_.__name__
    else:
        path = classpath(type_)
    return "<{}>".format(path)


@type_spec.register(str)
def type_spec_str(s):
    return "<{}>".format(s)


def parser_constructor_for_collection(cls):
    if is_named_tuple_class(cls):
        return UnStarred(cls)
    return get_constructor_for(cls)


byte_repr = "<0-255>"
regex_repr = "<regex>"
regex_bytes_repr = "<byte-regex>"
date_repr = "YYYY-MM-DD"
path_repr = "<path>"
binary_path_repr = "<binary-file>"
text_path_repr = "<text-file>"
classpath_type_repr = "path.to.type[params]"
classpath_function_repr = "path.to.function"
int_repr = type_spec(int)
float_repr = type_spec(float)
decimal_repr = "<decimal-str>"
fraction_repr = "{i}[/{i}]".format(i=int_repr)
complex_repr = "{f}[+{f}j]".format(f=float_repr)
range_repr = "{i}:{i}[:{i}]".format(i=int_repr)
datetime_repr = "YYYY-MM-DD[THH:MM:SS[.ms]]"
ipv4_repr = "<ipaddr>"
ipv6_repr = "<ipv6addr>"
url_repr = "scheme://netloc[/path][;params][?query][#fragment]"
uuid_repr = "[0-f]{32}"
ellipsis_ = "..."
any_repr = "____"


default_repr_values = {
    str: type_spec(str),
    int: int_repr,
    float: float_repr,
    complex: complex_repr,
    fractions.Fraction: fraction_repr,
    decimal.Decimal: decimal_repr,
    range: range_repr,
    datetime.date: date_repr,
    datetime.datetime: datetime_repr,
    pathlib.Path: path_repr,
    File: path_repr,
    BinaryFile: binary_path_repr,
    TextFile: text_path_repr,
    typing.Pattern: regex_repr,
    typing.Pattern[bytes]: regex_bytes_repr,
    ipaddress.IPv4Address: ipv4_repr,
    ipaddress.IPv6Address: ipv6_repr,
    URL: url_repr,
    uuid.UUID: uuid_repr,
    Empty: any_repr,
    typing.Callable: classpath_function_repr,
    types.FunctionType: classpath_function_repr,
    types.BuiltinFunctionType: classpath_function_repr,
    numbers.Number: "|".join(
        (int_repr, float_repr, complex_repr.replace("[", "").replace("]", ""))
    ),
    numbers.Real: "|".join((int_repr, float_repr)),
    numbers.Integral: int_repr,
}

default_value_repr_values = {
    sys.stdout: "stdout",
    sys.stderr: "stderr",
    sys.stdin: "stdin",
}


def repr_type(type_, supertype=None):
    if supertype is None:
        return classpath_type_repr
    return "{}<:{}".format(classpath_type_repr, parameterized_classpath(supertype))


def repr_value(value: object) -> str:
    """Represent a default value on the command line"""
    r = default_value_repr_values.get(value)
    if r is None:
        return str(value)
    return r


def maybe_map(f, it, exc=Exception):
    for i in it:
        try:
            res = f(i)
        except exc:
            pass
        else:
            yield res


def to_param_doc(param: Doc, name: str) -> Optional[str]:
    if param is None:
        return None
    elif isinstance(param, CallableDocs):
        p = param.params.get(name)
        if p:
            return p.doc
    elif isinstance(param, ParamDocs):
        p = param.get(name)
        if p:
            return p.doc
    elif isinstance(param, ParamDoc):
        return param.doc
    else:
        raise TypeError(
            "must pass introspection.CallableDocs/Params/Param for a docstring; got {}".format(
                type(param)
            )
        )


class Missing(list):
    # we make this a list subclass to allow argparse append actions to take place
    @classmethod
    def missing(cls, value):
        if isinstance(value, cls):
            # nothing appended yet
            return not value
        elif isinstance(value, type):
            return issubclass(value, cls)
        return False


class IODispatch(Dispatcher):
    """interject a custom exception into the dispatch process"""

    exc_type = IOError

    def __init__(self, name, return_type, doc=None):
        self.return_type = deconstruct_generic(return_type)
        super().__init__(name, doc=doc)

    def __call__(self, arg):
        try:
            return Dispatcher.__call__(self, arg)
        except Exception as e:
            raise self.exc_type(self.return_type, arg, e)


class TypeCheckPicklableWithType(PicklableWithType):
    def type_check(self, value):
        if not isinstance_generic(value, self.type_):
            raise self.exc_cls(self.type_, value)
        return value


class TypeCheckInput(TypeCheckPicklableWithType):
    exc_cls = TypedInputError
    decode = staticmethod(identity)

    def __call__(self, value):
        try:
            parsed = self.decode(value)
        except Exception as e:
            raise self.exc_cls(self.type_, value, e)
        else:
            return self.type_check(parsed)


class TypeCheckImport(TypeCheckInput):
    decode = staticmethod(import_object)

    def __call__(self, value):
        if isinstance(value, str):
            return super().__call__(value)
        else:
            return self.type_check(value)


class TypeCheckOutput(TypeCheckPicklableWithType):
    exc_cls = TypedOutputError
    encode = staticmethod(identity)

    def __init__(self, type_, *args, encoder=None):
        super().__init__((type_, *args))
        if encoder is not None:
            self.encode = encoder

    def __call__(self, value):
        value = self.type_check(value)
        try:
            out = self.encode(value)
        except Exception as e:
            raise self.exc_cls(self.type_, value, e)
        else:
            return out


class TypeCheckOutputFunc(TypeCheckOutput):
    encode = staticmethod(function_classpath)

    def type_check(self, value):
        if not callable(value):
            raise self.exc_cls(
                self.type_, value, TypeError("{} is not callable".format(value))
            )
        return value

    def __call__(self, value):
        path = super().__call__(value)
        if import_object(path) is not value:
            raise self.exc_cls(
                self.type_,
                value,
                ValueError(
                    "classpath {} does not refer to the same object as {}".format(
                        path, value
                    )
                ),
            )
        return path


class TypeCheckOutputType(TypeCheckOutput):
    encode = staticmethod(parameterized_classpath)



# coding:utf-8
import pytest
import re
from pathlib import Path
from datetime import date, datetime
from typing import (
    Mapping,
    List,
    MutableSet,
    Tuple,
    Union,
    Pattern,
    Counter,
    ByteString,
    ChainMap,
)
import collections as cl
from enum import Enum, Flag
from bourbaki.application.typed_io.config_decode import config_decoder


class SomeEnum(Enum):
    foo = 1
    bar = 2
    baz = 3


SomeFlag = Flag("SomeFlag", names="foo bar baz".split())


def same_value(a, b):
    assert a == b
    assert type(a) is type(b)


def same_contents(a, b):
    assert len(a) == len(b)
    assert type(a) is type(b)
    for a_, b_ in zip(a, b):
        same_value(a_, b_)


def same_keyvals(d, e):
    assert len(d) == len(d)
    assert type(d) is type(e)
    for k, l in zip(sorted(set(d)), sorted(set(e))):
        same_value(k, l)
        same_value(d[k], e[l])


@pytest.mark.parametrize(
    "type_,value,expected,cmp",
    [
        (int, 1, 1, same_value),
        (str, "foo", "foo", same_value),
        (type, "collections.OrderedDict", cl.OrderedDict, same_value),
        (range, "1:2:3", range(1, 2, 3), same_value),
        (range, [1, 2, 3], range(1, 2, 3), same_value),
        (SomeEnum, "foo", SomeEnum.foo, same_value),
        (SomeFlag, ["foo", "bar"], SomeFlag.foo | SomeFlag.bar, same_value),
        (
            Counter[str],
            {"foo": 1, "bar": 2},
            cl.Counter(["foo", "bar", "bar"]),
            same_keyvals,
        ),
        (
            Counter[int],
            [("1", "1"), ("2", "2"), ("3", "3")],
            cl.Counter({1: 1, 2: 2, 3: 3}),
            same_keyvals,
        ),
        (date, "2018-01-01", date(2018, 1, 1), same_value),
        (datetime, "2018-01-01T12:00:00.000", datetime(2018, 1, 1, 12), same_value),
        (bytes, [1, 2, 3], b"\x01\x02\x03", same_value),
        (ByteString, "b'foo'", b"foo", same_value),
        (Path, "foo/bar", Path("foo") / "bar", same_value),
        (List[bytes], ["b'foo'", [1, 2, 3]], [b"foo", b"\x01\x02\x03"], same_contents),
        (
            MutableSet[bytes],
            ["b'foo'", [1, 2, 3]],
            {b"foo", b"\x01\x02\x03"},
            same_contents,
        ),
        (
            Tuple[float, date, List[Path]],
            [1, "2018-01-01", ("foo", "bar")],
            (1.0, date(2018, 1, 1), [Path("foo"), Path("bar")]),
            same_contents,
        ),
        (Pattern[bytes], "foobar", re.compile(b"foobar"), same_value),
        (
            List[Tuple[float, bool]],
            ([1, True], [2, False]),
            [(1.0, True), (2.0, False)],
            same_contents,
        ),
        (
            Union[List[Tuple[date, tuple]], Mapping[date, tuple]],
            [["2018-01-01", [1, True]], ["2019-01-01", [False, 2]]],
            [(date(2018, 1, 1), (1, True)), (date(2019, 1, 1), (False, 2))],
            same_contents,
        ),
        (
            Union[Mapping[date, tuple], List[Tuple[date, ...]]],
            [["2018-01-01", [1, True]], ["2019-01-01", [False, 2]]],
            dict([(date(2018, 1, 1), (1, True)), (date(2019, 1, 1), (False, 2))]),
            same_contents,
        ),
        (Tuple[complex, ...], ["1+2j", 3 + 4j], (1 + 2j, 3 + 4j), same_contents),
        (
            ChainMap[int, range],
            [{1: "1:2", 2: [3, 4, 5]}, {"3": "6:7"}],
            cl.ChainMap({1: range(1, 2), 2: range(3, 4, 5), 3: range(6, 7)}),
            same_keyvals,
        ),
        (
            ChainMap[int, range],
            {1: "1:2", 2: [3, 4, 5], "3": "6:7"},
            cl.ChainMap({1: range(1, 2), 2: range(3, 4, 5), 3: range(6, 7)}),
            same_keyvals,
        ),
    ],
)
def test_postproc(type_, value, expected, cmp):
    postproc = config_decoder(type_)
    out = postproc(value)
    cmp(expected, out)



# coding:utf-8
from typing import Mapping, Any
import os
import functools
import io
import tempfile
import pytest
from itertools import chain, repeat
from copy import deepcopy
from sklearn.base import BaseEstimator
from multipledispatch import dispatch
from bourbaki.application.config import (
    load_config,
    dump_config,
    allow_unsafe_yaml,
    LEGAL_CONFIG_EXTENSIONS,
)
from bourbaki.application.typed_io.inflation import CLASSPATH_KEY, KWARGS_KEY
from bourbaki.application.typed_io.config_decode import config_decoder

# remove duplicate .yaml -> .yml
LEGAL_CONFIG_EXTENSIONS = tuple(e for e in LEGAL_CONFIG_EXTENSIONS if e != ".yaml")
NON_JSON_INI_EXTENSIONS = tuple(
    e for e in LEGAL_CONFIG_EXTENSIONS if e not in (".toml", ".json", ".ini")
)
NON_JSON_EXTENSIONS = tuple(
    e for e in LEGAL_CONFIG_EXTENSIONS if e not in (".toml", ".json")
)
NON_INI_EXTENSIONS = tuple(e for e in LEGAL_CONFIG_EXTENSIONS if e not in (".ini",))
ALLOW_INT_KEYS_EXTENSIONS = (".yml", ".py")

construct_instances_recursively = config_decoder(Mapping[str, BaseEstimator])

sklearnconf = {
    "forest": {
        CLASSPATH_KEY: "sklearn.ensemble.RandomForestClassifier",
        KWARGS_KEY: {"min_samples_split": 5, "n_estimators": 100},
    },
    "isolation_forest_n192_s1024": {
        CLASSPATH_KEY: "sklearn.ensemble.iforest.IsolationForest",
        KWARGS_KEY: {
            "bootstrap": True,
            "max_samples": 1024,
            "n_estimators": 192,
            "n_jobs": 7,
        },
    },
    "logistic_l1": {
        CLASSPATH_KEY: "sklearn.linear_model.LogisticRegression",
        KWARGS_KEY: {"penalty": "l1"},
    },
    "logistic_l1_highreg": {
        CLASSPATH_KEY: "sklearn.linear_model.LogisticRegression",
        KWARGS_KEY: {
            "C": 0.1,
            # 'class_weight': {0: 0.01, 1: 0.99},
            "penalty": "l1",
        },
    },
    "logistic_l2": {
        CLASSPATH_KEY: "sklearn.linear_model.LogisticRegression",
        KWARGS_KEY: {"penalty": "l2"},
    },
    "logistic_l2_highreg": {
        CLASSPATH_KEY: "sklearn.linear_model.LogisticRegression",
        KWARGS_KEY: {
            "C": 0.1,
            # 'class_weight': {0: 0.01, 1: 0.99},
            "penalty": "l2",
        },
    },
    "svc": {
        CLASSPATH_KEY: "sklearn.svm.SVC",
        KWARGS_KEY: {
            "C": 1.0,
            "cache_size": 1024,
            # 'class_weight': {0: 0.01, 1: 0.99},
            "coef0": 0.0,
            "degree": 2,
            "gamma": "auto",
            "kernel": "poly",
            "probability": True,
        },
    },
}

sklearnconf_w_int_keys = deepcopy(sklearnconf)
sklearnconf_w_int_keys["svc"][KWARGS_KEY]["class_weight"] = {0: 0.01, 1: 0.99}


fooconf = dict(foo="bar", baz=(1, 2, {3: 4, 5: [6, 7]}), qux=["foo", ("bar", "baz")])

# replace lists with tuples
@dispatch(dict)
def jsonify(obj, str_keys=True):
    f = str if str_keys else lambda x: x
    return {f(k): jsonify(v, str_keys=str_keys) for k, v in obj.items()}


@dispatch((list, tuple))
def jsonify(obj, str_keys=True):
    return list(map(functools.partial(jsonify, str_keys=str_keys), obj))


@dispatch(object)
def jsonify(obj, str_keys=True):
    return obj


@pytest.fixture(scope="function")
def tmp():
    f = tempfile.mktemp()
    yield f


def test_top_level_in_ini():
    # values with no sections dump and load in .ini as we would expect in .toml
    s = """
    value1 = "foo"
    [section]
    value2 = 1
    """
    f = io.StringIO(s)
    c = load_config(f, ext=".toml")
    newf = io.StringIO()
    dump_config(c, newf, ext=".ini")
    newf.seek(0)
    assert c == load_config(newf, ext=".ini")


def test_inflate_instances():
    models = construct_instances_recursively(sklearnconf_w_int_keys)
    for v in models.values():
        assert isinstance(v, BaseEstimator)


@pytest.mark.parametrize("ext", NON_JSON_INI_EXTENSIONS)
@pytest.mark.parametrize("conf", [sklearnconf, fooconf])
def test_dump_load_python_yaml(conf, ext, tmp):
    if not os.path.exists(tmp):
        with open(tmp + ext, "w"):
            pass
    dump_config(conf, tmp, ext=ext)
    conf_ = load_config(tmp + ext, disambiguate=True)
    assert jsonify(conf, str_keys=False) == jsonify(conf_, str_keys=False)
    os.remove(tmp + ext)


@pytest.mark.parametrize("conf", [sklearnconf, fooconf])
def test_dump_load_json(conf, tmp):
    ext = ".json"
    dump_config(conf, tmp, ext=ext)
    conf_ = load_config(tmp, disambiguate=True)
    assert jsonify(conf) == conf_
    # os.remove(tmp + ext)


@pytest.mark.parametrize(
    "ext,conf",
    chain(
        zip(NON_INI_EXTENSIONS, repeat(sklearnconf)),
        zip(ALLOW_INT_KEYS_EXTENSIONS, repeat(sklearnconf_w_int_keys)),
    ),
)
def test_dump_load_class_instances(ext, conf, tmp):
    dump_config(conf, tmp, ext=ext)
    conf_ = load_config(tmp, disambiguate=True)
    models = construct_instances_recursively(conf)
    models_ = construct_instances_recursively(conf_)

    for m in (models, models_):
        for v in m.values():
            assert isinstance(v, BaseEstimator)

    assert models_.keys() == models.keys()

    for k in models:
        m1 = models[k]
        m2 = models_[k]
        c = conf[k]
        for attr in c[KWARGS_KEY]:
            attr1 = getattr(m1, attr)
            if ext == ".json":
                attr1 = jsonify(attr1)

            assert attr1 == getattr(m2, attr)



#!/usr/bin/env python
# coding:utf-8
from typing import *
from typing import ChainMap
import collections
import itertools
from numbers import Number
import os
import tempfile

import pytest

from bourbaki.introspection.types import LazyType, get_generic_args
from bourbaki.introspection.types.compat import NEW_TYPING
from bourbaki.introspection.callables import call_repr
from bourbaki.application.typed_io import TypedIO
from bourbaki.application.typed_io.utils import *
from bourbaki.application.typed_io.exceptions import *
from bourbaki.application.typed_io.cli_repr_ import bool_cli_repr, KEY_VAL_JOIN_CHAR
from bourbaki.application.typed_io.utils import byte_repr, READ_MODES, WRITE_MODES
from bourbaki.application.typed_io.config_repr_ import (
    config_repr,
    bool_config_repr,
    bytes_config_repr,
)
from bourbaki.application.typed_io.config_encode import config_encoder
from bourbaki.application.typed_io.inflation import CLASSPATH_KEY, ARGS_KEY, KWARGS_KEY
from bourbaki.application.typed_io.cli_parse import cli_parser, cli_nargs, cli_repr

undefined = object()
skip = object()


def equal_contents(coll1, coll2):
    if type(coll1) is not type(coll2):
        return False
    if len(coll1) != len(coll2):
        return False
    return set(coll1) == set(coll2)


def equal_types(val1, val2):
    if type(val1) is not type(val2):
        return False
    return val1 == val2


class _TestCase:
    def __init__(
        self,
        type_,
        cli_nargs=undefined,
        cli_repr=undefined,
        config_repr=undefined,
        test_val=undefined,
        cli_test_val=undefined,
        config_test_val=undefined,
        encoded_eq=equal_types,
        multi_test=False,
    ):
        self.typed_io = TypedIO(type_)
        self.type_ = type_
        self.cli_nargs = cli_nargs
        self.cli_repr = cli_repr
        self.cli_test_val = cli_test_val
        self.config_repr = config_repr
        self.config_test_val = config_test_val
        self.test_val = test_val
        self.encoded_eq = encoded_eq
        self.multi_test = multi_test

    def __str__(self):
        return str(self.type_)

    def __repr__(self):
        return repr(self.type_)

    def should_test(self, attr):
        attr_ = dict(
            cli_parser="cli_test_val",
            config_decoder="config_test_val",
            config_encoder="test_val",
        ).get(attr, attr)
        selfattr = getattr(self, attr_)
        if selfattr is undefined:
            return False
        elif selfattr is skip:
            return None
        return True

    def _test_attr(self, attr):
        if self.should_test(attr):
            target, test = getattr(self.typed_io, attr), getattr(self, attr)
            assert target == test

    def _test_callable_attr(self, attr, target, test, exc, encoded=False):
        cond = self.should_test(attr)
        if cond:
            func = getattr(self.typed_io, attr)
            if not self.multi_test:
                test = [test]
            print(attr, "\n test: ", test[0], "\n target: ", target, file=sys.stderr)
            for test_case in test:
                if encoded:
                    cmp = self.encoded_eq
                    assert cmp(target, func(test_case))
                else:
                    assert target == func(test_case)
        elif cond is False:
            with pytest.raises(exc):
                func = getattr(self.typed_io, attr)

    def test_cli_parser(self):
        self._test_callable_attr(
            "cli_parser", self.test_val, self.cli_test_val, CLIIOUndefined
        )

    def test_config_decoder(self):
        self._test_callable_attr(
            "config_decoder", self.test_val, self.config_test_val, ConfigIOUndefined
        )

    def test_config_encoder(self):
        target_val = (
            self.config_test_val[0] if self.multi_test else self.config_test_val
        )
        test_val = [self.test_val] if self.multi_test else self.test_val
        self._test_callable_attr(
            "config_encoder", target_val, test_val, ConfigIOUndefined, encoded=True
        )

    def test_cli_nargs(self):
        self._test_attr("cli_nargs")

    def test_cli_repr(self):
        self._test_attr("cli_repr")

    def test_config_repr(self):
        self._test_attr("config_repr")


def mapping_cli_repr_str(keystr, valstr):
    return "{}{}{}".format(keystr, KEY_VAL_JOIN_CHAR, valstr)


tuple_str_ = ("foo", "bar", "baz")
list_str = list(tuple_str_)
set_str = set(tuple_str_)
frozenset_str = frozenset(tuple_str_)

tuple_int_ = tuple_int_int_int = (-1, 135514356, 0)
list_int_as_str = list(map(str, tuple_int_))
list_int = list(tuple_int_)
set_int = set(tuple_int_)
frozenset_int = frozenset(tuple_int_)
LazyListInt = LazyType["List[int]"]

cli_nargs.register(LazyListInt, as_const=True)(ZERO_OR_MORE)

bytes_cli_repr = byte_repr
bytes_int = [0, 1, 255]
bytes_str = ["0", "1", "255"]
bytes_ = bytes(bytes_int)
bytearray_ = bytearray(bytes_int)

any_coll_cli_repr = any_repr  # seq_cli_repr_template.format(t=any_repr)
any_coll_config_repr = [any_repr, ellipsis_]
int_coll_cli_repr = type_spec(int)  # seq_cli_repr_template.format(t=type_spec(int))
int_coll_config_repr = [type_spec(int), ellipsis_]

date_ = datetime.date.today()
date_epoch = datetime.date(1970, 1, 1)
date_str = date_.isoformat()
date_tup = [date_.year, date_.month, date_.day]

datetime_ = datetime.datetime.now()
datetime_str = datetime_.isoformat()
date_int = datetime_int = datetime_.timestamp()
datetime_tup = [
    datetime_.year,
    datetime_.month,
    datetime_.day,
    datetime_.hour,
    datetime_.minute,
    datetime_.second,
    datetime_.microsecond,
]

ipv4_str = "1.23.45.67"
ipv6_str = "12:34:a:b:c:d:e:f"
ipv4 = ipaddress.IPv4Address(ipv4_str)
ipv6 = ipaddress.IPv6Address(ipv6_str)
ipv4_int = int(ipv4)

map_int_tup_date_ip = {
    0: (date_, ipv4),
    -1234567: (date_ - datetime.timedelta(days=100000), ipv4),
}
map_int_or_date_tup_any_ip_ = {
    -1234567: (ipv4, ipv6),
    date_: (ipv6, ipv6),
    date_ - datetime.timedelta(days=100000): (ipv4, ipv4, ipv6, ipv6),
}
map_int_or_date_tup_any_ip_config = {
    str(k) if isinstance(k, int) else k.isoformat(): list(map(str, v))
    for k, v in map_int_or_date_tup_any_ip_.items()
}

any_ip = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
any_ip_config_repr = " OR ".join([ipv4_repr, ipv6_repr])

bool_float_or_str = Union[bool, float, str]
bool_float_or_str_cli_repr = "|".join((bool_cli_repr, type_spec(float), type_spec(str)))
bool_float_or_str_config_repr = " OR ".join(
    [bool_config_repr, type_spec(float), type_spec(str)]
)


class SimpleEQ:
    args = ()

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    @property
    def kwargs(self):
        kwargs = dict(self.__dict__)
        sig = signature(self.__init__)
        to_pop = [k for k in kwargs if k not in sig.parameters]
        for k in to_pop:
            kwargs.pop(k)
        kwargs.pop("args", None)
        return kwargs

    def __str__(self):
        args = self.args
        kwargs = self.kwargs
        return call_repr(type(self), args, kwargs)

    __repr__ = __str__


class ConfigReprMeta(type):
    @property
    def __classpath__(self):
        return parameterized_classpath(self)

    __sig_args__ = None
    __kwargs__ = {}

    @property
    def config_repr(self):
        repr_ = {CLASSPATH_KEY: self.__classpath__, KWARGS_KEY: self.__kwargs__}
        if self.__sig_args__ is not None:
            repr_[ARGS_KEY] = self.__sig_args__
        return repr_


class SimpleConfigRepr(SimpleEQ, metaclass=ConfigReprMeta):
    def config_encode(self, *args):
        ns = dict(self.__dict__)
        ns.pop("args", None)
        repr_ = {CLASSPATH_KEY: classpath(type(self)), KWARGS_KEY: ns}
        if hasattr(self, "args"):
            repr_[ARGS_KEY] = list(self.args)
        return repr_


@config_encoder.register(SimpleConfigRepr)
def simple_config_encode(cls, *args):
    def encode(val):
        return val.config_encode(*args)

    return encode


class custom_class(SimpleConfigRepr):
    def __init__(self):
        pass


class custom_class_with_varargs(SimpleConfigRepr):
    def __init__(self, *args: datetime.datetime):
        self.args = args

    __sig_args__ = [datetime_repr, ellipsis_]


class custom_class_with_kwargs(SimpleConfigRepr):
    def __init__(self, **kwargs: Union[int, str]):
        self.__dict__.update(kwargs)

    __kwargs__ = {ellipsis_: " OR ".join((type_spec(int), type_spec(str)))}


T_co = TypeVar("T", covariant=True)


if NEW_TYPING:
    _SimpleConfigReprGenericMeta_bases = ()
else:
    _SimpleConfigReprGenericMeta_bases = (GenericMeta,)


class SimpleConfigReprGenericMeta(ConfigReprMeta, *_SimpleConfigReprGenericMeta_bases):
    @property
    def __kwargs__(cls):
        args = get_generic_args(cls)
        if not args:
            return {"xs": {type_spec(int): any_repr, ellipsis_: ellipsis_}}
        return {"xs": {type_spec(int): config_repr(args[0]), ellipsis_: ellipsis_}}

    @property
    def __sig_args__(cls):
        args = get_generic_args(cls)
        if not args:
            return [any_repr, ellipsis_]
        return [config_repr(args[0]), ellipsis_]


class FooTup(NamedTuple):
    foo: int
    bar: str


class custom_generic_class(
    SimpleConfigRepr, Generic[T_co], metaclass=SimpleConfigReprGenericMeta
):
    def __init__(self, *args: T_co, xs: Mapping[int, T_co]):
        self.xs = xs
        self.args = args

    def config_encode(self, t=None):
        conf = super().config_encode()
        if t is None:
            enc = identity
        else:
            enc = config_encoder(t)
            conf[CLASSPATH_KEY] = conf[CLASSPATH_KEY] + "[{}]".format(
                parameterized_classpath(t)
            )

        conf["__kwargs__"]["xs"] = {
            str(k): enc(v) for k, v in conf["__kwargs__"]["xs"].items()
        }
        conf["__args__"] = [enc(v) for v in conf["__args__"]]
        return conf


if NEW_TYPING:
    # can't mix metaclasses with Generic in python3.7! No worry, this only complicates testing;
    # the class properties defined above aren't necessary in real use cases
    for t, attr in itertools.product(
        (str, datetime.date),
        ("__kwargs__", "__classpath__", "__sig_args__", "config_repr"),
    ):
        gen = custom_generic_class[t]
        gen.__dict__[attr] = getattr(SimpleConfigReprGenericMeta, attr).fget(gen)


test_cases = [
    # atomic
    _TestCase(
        bool,
        None,
        bool_cli_repr,
        bool_config_repr,
        True,
        ["true", "True"],
        [True],
        multi_test=True,
    ),
    _TestCase(int, None, type_spec(int), type_spec(int), -1234, "-1234", -1234),
    _TestCase(
        datetime.date,
        None,
        date_repr,
        date_repr,
        date_,
        [date_str],
        [date_str, date_int, date_tup],
        multi_test=True,
    ),
    _TestCase(
        datetime.datetime,
        None,
        datetime_repr,
        datetime_repr,
        datetime_,
        datetime_str,
        datetime_str,
    ),
    _TestCase(
        ipaddress.IPv4Address, None, ipv4_repr, ipv4_repr, ipv4, ipv4_str, ipv4_str
    ),
    _TestCase(
        ipaddress.IPv6Address, None, ipv6_repr, ipv6_repr, ipv6, ipv6_str, ipv6_str
    ),
    # types and functions
    _TestCase(
        Type[Number],
        None,
        classpath_type_repr + "<:numbers.Number",
        classpath_type_repr + "<:numbers.Number",
        test_val=complex,
        cli_test_val="builtins.complex",
        config_test_val="complex",
    ),
    _TestCase(
        types.FunctionType,
        None,
        classpath_function_repr,
        classpath_function_repr,
        test_val=equal_contents,
        cli_test_val="{}.equal_contents".format(__name__),
        config_test_val="{}.equal_contents".format(__name__),
    ),
    # Unions
    _TestCase(
        any_ip,
        None,
        "{}|{}".format(ipv4_repr, ipv6_repr),
        any_ip_config_repr,
        ipv4,
        ipv4_str,
        ipv4_str,
    ),
    _TestCase(
        any_ip,
        None,
        "{}|{}".format(ipv4_repr, ipv6_repr),
        any_ip_config_repr,
        ipv6,
        ipv6_str,
        ipv6_str,
    ),
    _TestCase(
        bool_float_or_str,
        None,
        bool_float_or_str_cli_repr,
        bool_float_or_str_config_repr,
        "asdf",
        "asdf",
        "asdf",
    ),
    _TestCase(
        bool_float_or_str,
        None,
        bool_float_or_str_cli_repr,
        bool_float_or_str_config_repr,
        3.14,
        ".314e1",
        3.14,
    ),
    _TestCase(
        bool_float_or_str,
        None,
        bool_float_or_str_cli_repr,
        bool_float_or_str_config_repr,
        False,
        "FALSE",
        False,
    ),
    # collections
    _TestCase(
        bytes,
        ZERO_OR_MORE,
        bytes_cli_repr,
        bytes_config_repr,
        bytes_,
        bytes_str,
        bytes_int,
    ),
    _TestCase(
        bytearray,
        ZERO_OR_MORE,
        bytes_cli_repr,
        bytes_config_repr,
        bytearray_,
        bytes_str,
        bytes_int,
    ),
    _TestCase(
        tuple,
        ZERO_OR_MORE,
        any_coll_cli_repr,
        any_coll_config_repr,
        tuple_str_,
        list_str,
        list_str,
    ),
    _TestCase(
        set,
        ZERO_OR_MORE,
        any_coll_cli_repr,
        any_coll_config_repr,
        set_str,
        list_str,
        list_str,
        equal_contents,
    ),
    _TestCase(
        frozenset,
        ZERO_OR_MORE,
        any_coll_cli_repr,
        any_coll_config_repr,
        frozenset_str,
        list_str,
        list_str,
        equal_contents,
    ),
    _TestCase(
        list,
        ZERO_OR_MORE,
        any_coll_cli_repr,
        any_coll_config_repr,
        list_str,
        list_str,
        list_str,
    ),
    # mappings
    _TestCase(
        Mapping[int, Tuple[datetime.date, ipaddress.IPv4Address]],
        undefined,
        undefined,
        {type_spec(int): [date_repr, ipv4_repr], ellipsis_: ellipsis_},
        test_val=map_int_tup_date_ip,
        cli_test_val=undefined,
        config_test_val={
            str(k): [d.isoformat(), str(i)] for k, (d, i) in map_int_tup_date_ip.items()
        },
    ),
    _TestCase(
        Mapping[
            Union[int, datetime.date],
            Tuple[Union[ipaddress.IPv4Address, ipaddress.IPv6Address], ...],
        ],
        undefined,
        undefined,
        config_repr={
            "{} OR {}".format(type_spec(int), date_repr): [
                any_ip_config_repr,
                ellipsis_,
            ],
            ellipsis_: ellipsis_,
        },
        test_val=map_int_or_date_tup_any_ip_,
        cli_test_val=undefined,
        config_test_val=map_int_or_date_tup_any_ip_config,
    ),
    _TestCase(
        ChainMap[Type, datetime.date],
        ZERO_OR_MORE,
        cli_repr=mapping_cli_repr_str(classpath_type_repr, date_repr),
        config_repr={classpath_type_repr: date_repr, ellipsis_: ellipsis_},
        test_val=collections.ChainMap({int: date_, complex: date_epoch}),
        cli_test_val=[
            "{}={}".format("int", date_.isoformat()),
            "{}={}".format("builtins.complex", date_epoch.isoformat()),
        ],
        config_test_val=[{"int": date_.isoformat(), "complex": date_epoch.isoformat()}],
    ),
    _TestCase(
        Counter[Union[types.FunctionType, types.BuiltinFunctionType]],
        ZERO_OR_MORE,
        cli_repr=mapping_cli_repr_str(classpath_function_repr, type_spec(int)),
        config_repr={classpath_function_repr: type_spec(int), ellipsis_: ellipsis_},
        test_val={len: 2, __import__: -1234567890},
        cli_test_val=[
            "{}={}".format("len", 2),
            "{}={}".format("__import__", -1234567890),
        ],
        config_test_val={"len": 2, "__import__": -1234567890},
    ),
    # parameterized collections
    _TestCase(
        List[int],
        ZERO_OR_MORE,
        int_coll_cli_repr,
        int_coll_config_repr,
        list_int,
        list_int_as_str,
        list_int,
    ),
    _TestCase(
        Set[int],
        ZERO_OR_MORE,
        int_coll_cli_repr,
        int_coll_config_repr,
        set_int,
        list_int_as_str,
        list_int,
        equal_contents,
    ),
    _TestCase(
        FrozenSet[int],
        ZERO_OR_MORE,
        int_coll_cli_repr,
        int_coll_config_repr,
        frozenset_int,
        list_int_as_str,
        list_int,
        equal_contents,
    ),
    _TestCase(
        Tuple[int, ...],
        ZERO_OR_MORE,
        int_coll_cli_repr,
        int_coll_config_repr,
        tuple_int_,
        list_int_as_str,
        list_int,
    ),
    # tuples
    _TestCase(
        Tuple[int, int, int],
        3,
        tuple([type_spec(int)] * 3),
        [type_spec(int)] * 3,
        tuple_int_,
        list_int_as_str,
        list_int,
    ),
    _TestCase(
        Tuple[int],
        1,
        (type_spec(int),),
        [type_spec(int)],
        tuple_int_[:1],
        list_int_as_str[:1],
        list_int[:1],
    ),
    _TestCase(
        Tuple[bool, datetime.datetime, ipaddress.IPv4Address],
        3,
        (bool_cli_repr, datetime_repr, ipv4_repr),
        [bool_config_repr, datetime_repr, ipv4_repr],
        (False, datetime_, ipv4),
        [["false", datetime_str, ipv4_str]],
        [[False, datetime_str, ipv4_str], [False, datetime_int, ipv4_int]],
        multi_test=True,
    ),
    _TestCase(
        FooTup,
        2,
        (type_spec(int), type_spec(str)),
        {"foo": type_spec(int), "bar": type_spec(str)},
        FooTup(1, "2"),
        [["1", 2]],
        [{"foo": 1, "bar": "2"}, [1, "2"]],
        multi_test=True,
    ),
    # Lazy types
    # type_spec here rather than the usual cli_repr because we don't want to load the type to print CLI help
    _TestCase(
        LazyType["datetime.datetime"],
        None,
        type_spec(datetime.datetime),
        datetime_repr,
        datetime_,
        datetime_str,
        datetime_str,
    ),
    _TestCase(
        LazyListInt,
        ZERO_OR_MORE,
        "<List[int]>",
        [type_spec(int), ellipsis_],
        list_int,
        list_int_as_str,
        list_int,
    ),  # we registered with cli_nargs above to get '*' rather than None
    # custom classes
    _TestCase(
        custom_class,
        undefined,
        undefined,
        custom_class.config_repr,
        test_val=custom_class(),
        cli_test_val=undefined,
        config_test_val=custom_class().config_encode(),
    ),
    _TestCase(
        custom_class_with_varargs,
        undefined,
        undefined,
        config_repr=custom_class_with_varargs.config_repr,
        test_val=custom_class_with_varargs(datetime_),
        cli_test_val=undefined,
        config_test_val=custom_class_with_varargs(datetime_).config_encode(),
    ),
    _TestCase(
        custom_class_with_kwargs,
        undefined,
        undefined,
        config_repr=custom_class_with_kwargs.config_repr,
        test_val=custom_class_with_kwargs(a=1, b="two"),
        cli_test_val=undefined,
        config_test_val=custom_class_with_kwargs(a=1, b="two").config_encode(),
    ),
    # custom generic classes
    _TestCase(
        custom_generic_class,
        undefined,
        undefined,
        config_repr=custom_generic_class.config_repr,
        test_val=custom_generic_class(1, 2, 3, xs={1: "two", 3: "four"}),
        cli_test_val=undefined,
        config_test_val=custom_generic_class(
            1, 2, 3, xs={1: "two", 3: "four"}
        ).config_encode(),
    ),
    _TestCase(
        custom_generic_class[str],
        undefined,
        undefined,
        config_repr=custom_generic_class[str].config_repr,
        test_val=custom_generic_class("1", "2", "3", xs={1: "two", 3: "four"}),
        cli_test_val=undefined,
        config_test_val=custom_generic_class(
            "1", "2", "3", xs={"1": "two", "3": "four"}
        ).config_encode(str),
    ),
    _TestCase(
        custom_generic_class[datetime.date],
        undefined,
        undefined,
        config_repr=custom_generic_class[datetime.date].config_repr,
        test_val=custom_generic_class(date_, xs={1: date_}),
        cli_test_val=undefined,
        config_test_val=custom_generic_class(date_, xs={"1": date_}).config_encode(
            datetime.date
        ),
    ),
]


@pytest.mark.parametrize("test_case", test_cases)
def test_cli_nargs(test_case: _TestCase):
    test_case.test_cli_nargs()


@pytest.mark.parametrize("test_case", test_cases)
def test_cli_repr(test_case: _TestCase):
    test_case.test_cli_repr()


@pytest.mark.parametrize("test_case", test_cases)
def test_cli_parser(test_case: _TestCase):
    test_case.test_cli_parser()


@pytest.mark.parametrize("test_case", test_cases)
def test_config_repr(test_case: _TestCase):
    test_case.test_config_repr()


@pytest.mark.parametrize("test_case", test_cases)
def test_config_decoder(test_case: _TestCase):
    test_case.test_config_decoder()


@pytest.mark.parametrize("test_case", test_cases)
def test_config_encoder(test_case: _TestCase):
    test_case.test_config_encoder()


@pytest.mark.parametrize(
    "fileclass1,fileclass2",
    [
        (File, File),
        (File["w"], File),
        (File["r"], File),
        (File["wb"], File),
        (File["rb"], File),
        (BinaryFile, File),
        (TextFile, File),
        (File["w"], TextFile),
        (File["r"], TextFile),
        (File["wb"], BinaryFile),
        (File["rb"], BinaryFile),
        (File["rb"], io.BufferedReader),
        (File["wb"], io.BufferedWriter),
        (File["rb+"], io.BufferedRandom),
        (File["r", "utf-8"], io.TextIOWrapper),
        (File["w", "ascii"], io.TextIOWrapper),
        (File["r+", "utf16"], io.TextIOWrapper),
    ],
)
def test_file_issubclass(fileclass1, fileclass2):
    assert issubclass(fileclass1, fileclass2)
    if fileclass1 != fileclass2:
        assert not issubclass(fileclass2, fileclass1)


@pytest.mark.parametrize(
    "file,fileclass", [(io.BytesIO(), File["wb"]), (io.StringIO(), File["w+"])]
)
def test_file_isinstance(file, fileclass):
    assert isinstance(file, fileclass)


@pytest.mark.parametrize("mode", list(READ_MODES))
def test_open_file_isinstance_by_read_mode(mode):
    path = tempfile.mktemp()
    with open(path, "w") as f:
        # create file
        pass
    with open(path, mode) as f:
        assert isinstance(f, File[mode])
    assert isinstance(f, File[mode])
    os.remove(path)


@pytest.mark.parametrize("mode", list(WRITE_MODES))
def test_open_file_isinstance_by_write_mode(mode):
    path = tempfile.mktemp()
    if "x" not in mode:
        with open(path, "w") as f:
            # create file
            pass
    with open(path, mode) as f:
        assert isinstance(f, File[mode])
    assert isinstance(f, File[mode])
    os.remove(path)



# coding:utf-8
import typing
import types
import ast
import collections
import decimal
import enum
import fractions
import pathlib
import ipaddress
import datetime
import uuid
from urllib.parse import ParseResult as URL, urlparse
from functools import partial
from bourbaki.introspection.callables import UnStarred
from bourbaki.introspection.imports import import_object
from bourbaki.introspection.generic_dispatch import (
    GenericTypeLevelSingleDispatch,
    UnknownSignature,
)
from bourbaki.introspection.types import (
    issubclass_generic,
    get_constructor_for,
    NamedTupleABC,
    NonStrCollection,
    NonAnyStrCollection,
    LazyType,
)
from bourbaki.introspection.generic_dispatch_helpers import (
    LazyWrapper,
    CollectionWrapper,
    TupleWrapper,
    MappingWrapper,
    NamedTupleWrapper,
    UnionWrapper,
)
from bourbaki.application.typed_io.inflation import inflate_config
from .parsers import (
    parse_regex_bytes,
    parse_regex,
    parse_range,
    parse_iso_date,
    parse_iso_datetime,
    parse_bool,
    parse_path,
    EnumParser,
    FlagParser,
)
from .exceptions import (
    ConfigTypedInputError,
    ConfigIOUndefined,
    ConfigUnionInputError,
    ConfigCollectionKeysNotAllowed,
    ConfigCallableInputError,
)
from .utils import identity, Empty, IODispatch, TypeCheckInput, PicklableWithType, File
from .parsers import TypeCheckImportFunc, TypeCheckImportType
from .config_repr_ import bytes_config_key_repr


class ConfigDecodeDispatch(IODispatch):
    exc_type = ConfigTypedInputError


# basic decoders

config_parse_bool = partial(parse_bool, exc_type=ConfigTypedInputError)

to_int = ConfigDecodeDispatch("to_int", int)
to_int.register(int)(identity)
to_int.register(str)(int)

to_bool = ConfigDecodeDispatch("to_bool", bool)
to_bool.register(bool)(identity)
to_bool.register(str)(config_parse_bool)

to_float = ConfigDecodeDispatch("to_float", float)
to_float.register(float)(identity)
to_float.register((int, str))(float)

to_complex = ConfigDecodeDispatch("to_complex", complex)
to_complex.register(complex)(identity)
to_complex.register((int, str, float))(complex)

to_fraction = ConfigDecodeDispatch("to_fraction", fractions.Fraction)
to_fraction.register(fractions.Fraction)(identity)
to_fraction.register((int, str, float))(fractions.Fraction)


@to_fraction.register((list, tuple))
def numerator_denominator_to_fraction(tup):
    if len(tup) != 2:
        raise ValueError(
            "If instantiated from a tuple, Fraction requires a 2-tuple; got {}".format(
                tup
            )
        )
    return fractions.Fraction(*tup)


to_decimal = ConfigDecodeDispatch("to_decimal", decimal.Decimal)
to_decimal.register(decimal.Decimal)(identity)
to_decimal.register((int, str, float))(decimal.Decimal)

to_bytes = ConfigDecodeDispatch("to_bytes", bytes)
to_bytes.register(bytes)(identity)
to_bytes.register((list, bytearray))(bytes)


@to_bytes.register(str)
def str_to_bytes(s, type_=bytes, exc_type=ConfigTypedInputError):
    try:
        t = ast.parse(s, mode="eval")
    except SyntaxError as e:
        raise exc_type(type_, s, e)
    else:
        expr = t.body
        if not (isinstance(expr, ast.Constant) and isinstance(expr.value, bytes)):
            raise exc_type(
                type_,
                s,
                SyntaxError(
                    "{} is not a legal bytes expression; use format {}".format(
                        repr(s), bytes_config_key_repr
                    )
                ),
            )
        b = eval(s)
        if type_ is bytearray:
            b = bytearray(b)
        return b


to_bytearray = ConfigDecodeDispatch("to_bytearray", bytearray)
to_bytearray.register(bytearray)(identity)
to_bytearray.register((list, bytes))(bytearray)
to_bytearray.register(str)(partial(str_to_bytes, type_=bytearray))

to_date = ConfigDecodeDispatch("to_date", datetime.date)
to_date.register(datetime.date)(identity)
to_date.register((int, float))(datetime.date.fromtimestamp)
to_date.register(str)(parse_iso_date)


@to_date.register((tuple, list))
def date_from_tuple(t):
    return datetime.date(*t)


to_datetime = ConfigDecodeDispatch("to_datetime", datetime.datetime)
to_datetime.register(datetime.datetime)(identity)
to_datetime.register((int, float))(datetime.datetime.fromtimestamp)
to_datetime.register(str)(parse_iso_datetime)


@to_datetime.register((tuple, list))
def datetime_from_tuple(t):
    return datetime.datetime(*t)


to_range = ConfigDecodeDispatch("to_range", range)
to_range.register(str)(parse_range)


@to_range.register((list, tuple))
def range_from_tuple(tup):
    return range(*tup)


config_key_decoder_methods = {
    int: to_int,
    bool: to_bool,
    float: to_float,
    complex: to_complex,
    fractions.Fraction: to_fraction,
    decimal.Decimal: to_decimal,
    range: to_range,
    bytes: to_bytes,
    bytearray: to_bytearray,
    typing.ByteString: to_bytes,
    typing.Pattern: parse_regex,
    typing.Pattern[str]: parse_regex,
    typing.Pattern[bytes]: parse_regex_bytes,
    datetime.date: to_date,
    datetime.datetime: to_datetime,
    uuid.UUID: uuid.UUID,
    URL: urlparse,
    ipaddress.IPv4Address: ipaddress.IPv4Address,
    ipaddress.IPv6Address: ipaddress.IPv6Address,
    pathlib.Path: parse_path,
}


# The main dispatcher

config_decoder = GenericTypeLevelSingleDispatch(
    "config_decoder", isolated_bases=[typing.Union, typing.Generic]
)

config_key_decoder = GenericTypeLevelSingleDispatch(
    "config_key_decoder", isolated_bases=[typing.Union, typing.Generic]
)


# no typecheck for unannotated params
config_decoder.register(Empty, as_const=True)(inflate_config)

# File subclasses are their own parsers
config_decoder.register(File)(identity)
config_key_decoder.register(File)(identity)


# don't allow inflation for simple JSON-encodable atomic types
@config_decoder.register_all(str, type(None))
class TypeCheckConfig(TypeCheckInput):
    exc_cls = ConfigTypedInputError


# typecheck annotated params
@config_decoder.register_all(typing.Any, typing.Generic)
class TypeCheckInflateConfig(TypeCheckInput):
    exc_cls = ConfigTypedInputError

    def __init__(self, type_, *args):
        super().__init__(type_, *args)

    def decode(self, conf):
        # elide expensive construction if we can rule out that the classpath in the config is wrong from the start
        return inflate_config(conf, target_type=self.type_)


@config_decoder.register(typing.Callable)
class TypeCheckInflateCallableConfig(TypeCheckInflateConfig):
    exc_cls = ConfigCallableInputError

    def decode(self, conf):
        if isinstance(conf, str):
            # try a function import for a string
            return import_object(conf)
        return super().decode(conf)


@config_decoder.register_all(types.FunctionType, types.BuiltinFunctionType)
@config_key_decoder.register_all(types.FunctionType, types.BuiltinFunctionType)
class TypeCheckImportFuncConfig(TypeCheckImportFunc):
    exc_cls = ConfigTypedInputError


@config_decoder.register(typing.Type)
@config_key_decoder.register(typing.Type)
class TypeCheckImportTypeConfig(TypeCheckImportType):
    exc_cls = ConfigTypedInputError


# base for decoders that decode collections
class GenericConfigDecoderMixin(PicklableWithType):
    getter = config_decoder
    get_reducer = staticmethod(get_constructor_for)
    legal_container_types = None
    helper_cls = None
    exc_cls = ConfigTypedInputError
    init_type = True

    def __init__(self, generic, *args):
        if self.init_type:
            PicklableWithType.__init__(self, generic, *args)
        self.helper_cls.__init__(self, generic, *args)

    def typecheck(self, conf):
        if self.legal_container_types is None:
            return conf
        if not isinstance(conf, self.legal_container_types):
            raise ConfigTypedInputError(
                self.type_,
                conf,
                TypeError(
                    "{} is not an instance of any of {}".format(
                        conf, self.legal_container_types
                    )
                ),
            )
        return conf

    def __call__(self, conf):
        arg = self.typecheck(conf)
        try:
            return self.helper_cls.__call__(self, arg)
        except Exception as e:
            raise self.exc_cls(self.type_, conf, e)


@config_decoder.register(typing.Collection)
class CollectionConfigDecoder(GenericConfigDecoderMixin, CollectionWrapper):
    legal_container_types = (NonAnyStrCollection,)
    helper_cls = CollectionWrapper


# don't allow sequences to parse from unordered collections
@config_decoder.register(typing.Sequence)
class SequenceConfigDecoder(CollectionConfigDecoder):
    legal_container_types = (typing.Sequence,)


@config_decoder.register(typing.Mapping)
class MappingConfigDecoder(GenericConfigDecoderMixin, MappingWrapper):
    legal_container_types = (collections.abc.Mapping, NonAnyStrCollection)
    helper_cls = MappingWrapper

    def __init__(self, coll_type, key_type=typing.Any, val_type=Empty):
        if issubclass_generic(key_type, NonStrCollection):
            raise ConfigCollectionKeysNotAllowed((coll_type, key_type, val_type))
        super().__init__(coll_type, key_type, val_type)


@config_decoder.register(typing.ChainMap)
class ChainMapConfigDecoder(MappingConfigDecoder):
    reduce = dict

    def __call__(self, arg):
        to_map = partial(self.helper_cls.__call__, self)
        if isinstance(arg, typing.Sequence):
            maps = arg
        else:  # mapping
            maps = [arg]
        # these will typecheck
        maps = map(to_map, maps)
        try:
            return collections.ChainMap(*maps)
        except Exception as e:
            raise self.exc_cls(self.type_, arg, e)


@config_decoder.register(typing.Counter)
class CounterConfigEncoder(MappingConfigDecoder):
    init_type = False

    def __init__(self, coll_type, key_type):
        super().__init__(coll_type, key_type, int)
        # can't pass two type args to Counter, hence init_type = False above and we set the type_ attr manually here
        PicklableWithType.__init__(self, coll_type, key_type)

    def __call__(self, arg):
        # don't actually count entries; treat them as key-value tuples
        try:
            it = self.call_iter(arg)
        except Exception as e:
            raise self.exc_cls(self.type_, arg, e)
        return self.reduce(dict(it))


@config_decoder.register(typing.Tuple)
class TupleConfigDecoder(GenericConfigDecoderMixin, TupleWrapper):
    _collection_cls = SequenceConfigDecoder
    legal_container_types = (NonAnyStrCollection,)
    helper_cls = TupleWrapper


class _DictFromNamedTupleIter:
    def __init__(self, tuple_cls):
        self.tuple_cls = tuple_cls

    def __call__(self, keyvals):
        return self.tuple_cls(**dict(keyvals))


@config_decoder.register(NamedTupleABC)
class NamedTupleConfigDecoder(GenericConfigDecoderMixin, NamedTupleWrapper):
    get_named_reducer = staticmethod(_DictFromNamedTupleIter)
    get_reducer = staticmethod(UnStarred)
    legal_container_types = (typing.Sequence, typing.Mapping)
    helper_cls = NamedTupleWrapper


@config_decoder.register(typing.Union)
class UnionConfigDecoder(GenericConfigDecoderMixin, UnionWrapper):
    tolerate_errors = (ConfigIOUndefined, UnknownSignature)
    reduce = staticmethod(next)
    helper_cls = UnionWrapper
    exc_class = ConfigUnionInputError


@config_decoder.register(LazyType)
class LazyConfigDecoder(LazyWrapper):
    getter = config_decoder


# enums
@config_decoder.register_all(enum.Enum, enum.IntEnum)
@config_key_decoder.register_all(enum.Enum, enum.IntEnum)
def enum_config_decoder(enum_type):
    return EnumParser(enum_type).config_decode


@config_decoder.register_all(enum.Flag, enum.IntFlag)
@config_key_decoder.register_all(enum.Flag, enum.IntFlag)
def enum_config_decoder(enum_type):
    return FlagParser(enum_type).config_decode


# basic types that require parsing but not typechecking, because the parsers are well-typed
config_decoder.register_from_mapping(config_key_decoder_methods, as_const=True)
config_key_decoder.register_from_mapping(config_key_decoder_methods, as_const=True)

config_key_decoder.register(typing.Any)(identity)


@config_key_decoder.register(typing.Union)
class UnionConfigKeyDecoder(UnionConfigDecoder):
    getter = config_key_decoder



# coding: utf-8
from operator import attrgetter
import multiprocessing
import threading
import logging
import sys
import traceback
from getpass import getpass
import smtplib
import email.utils
from email.message import EmailMessage
from logging import StreamHandler, ERROR, LogRecord
from logging.handlers import SMTPHandler, MemoryHandler, RotatingFileHandler
from .config import validate_log_level_int


class BufferingSMTPHandler(MemoryHandler):
    def __init__(
        self,
        capacity,
        mailhost,
        toaddrs,
        subject=None,
        flushLevel=ERROR,
        *,
        credentials=None,
        fromaddr=None,
        secure=None,
        mailport=None,
        timeout=5.0
    ):
        flushLevel = validate_log_level_int(flushLevel)

        if isinstance(credentials, str):
            credentials = (
                credentials,
                getpass("Please enter a password for {}: ".format(credentials)),
            )

        if fromaddr is None:
            if not isinstance(credentials, (list, tuple)) or len(credentials) != 2:
                raise ValueError(
                    "you must supply either fromaddr or credentials=(uername, password); "
                    "fromaddr is None but credentials = {}".format(credentials)
                )
            fromaddr = credentials[0]

        if isinstance(toaddrs, str):
            toaddrs = [toaddrs]
        elif not toaddrs:
            raise ValueError(
                "you must supply toaddrs, either a single email address or a list thereof"
            )

        if mailport is not None:
            # SMTPHandler uses a tuple for this
            mailhost = (mailhost, mailport)
        elif not isinstance(mailhost, (list, tuple)) or len(mailhost) != 2:
            raise ValueError(
                "If mailport is not explicitly passed, mailhost must be a (host, port) tuple; got {}".format(
                    mailhost
                )
            )

        MemoryHandler.__init__(self, capacity, flushLevel=flushLevel)
        SMTPHandler.__init__(
            self,
            mailhost=mailhost,
            fromaddr=fromaddr,
            toaddrs=toaddrs,
            subject=subject,
            credentials=credentials,
            secure=secure,
            timeout=timeout,
        )

    def send_mail(self, content, subject=None):
        msg = EmailMessage()
        msg["From"] = self.fromaddr
        msg["To"] = ",".join(self.toaddrs)
        subject = subject or self.subject
        if subject is not None:
            msg["Subject"] = subject
        msg["Date"] = email.utils.localtime()
        msg.set_content(content)

        port = self.mailport or smtplib.SMTP_PORT
        smtp = smtplib.SMTP(self.mailhost, port, timeout=self.timeout)
        try:
            if self.username:
                if self.secure is not None:
                    smtp.ehlo()
                    smtp.starttls(*self.secure)
                    smtp.ehlo()
                smtp.login(self.username, self.password)
            smtp.send_message(msg)
        finally:
            smtp.quit()

    def get_content(self):
        return "\n".join(map(self.format, self.buffer))

    def get_subject(self):
        if self.subject:
            return self.subject
        top_record = max(self.buffer, key=attrgetter("levelno"))
        top_records = [r for r in self.buffer if r.levelno >= top_record.levelno]
        names = sorted(set(r.name for r in top_records))
        return "{} messages from loggers {}".format(
            top_record.levelname, ", ".join(names)
        )

    def flush(self):
        if len(self.buffer) > 0:
            content = self.get_content()
            subject = self.get_subject()

            try:
                self.send_mail(content, subject)
            except:
                self.handleError(
                    LogRecord(
                        self.name,
                        self.level,
                        pathname=None,
                        lineno=None,
                        msg=content,
                        args=(),
                        exc_info=sys.exc_info(),
                    )
                )  # no particular record
            finally:
                self.buffer = []

    def __del__(self):
        self.flush()


class MultiProcHandler(logging.Handler):
    """
    Adapted from zzzeek's answer at:
    https://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python
    Tweaked and subclassed here - added the get_subhandler() method for generality.
    """

    subhandler_cls = None

    def __init__(self, *args, **kwargs):
        logging.Handler.__init__(self)

        self._handler = self.get_subhandler(*args, **kwargs)
        self.queue = multiprocessing.Queue(-1)

        t = threading.Thread(target=self.receive, daemon=True)
        t.start()

    def get_subhandler(self, *args, **kwargs):
        return self.subhandler_cls(*args, **kwargs)

    def setFormatter(self, fmt):
        logging.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def receive(self):
        while True:
            try:
                record = self.queue.get()
                self._handler.emit(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except:
                traceback.print_exc(file=sys.stderr)

    def send(self, s):
        self.queue.put_nowait(s)

    def _format_record(self, record):
        # ensure that exc_info and args
        # have been stringified.  Removes any chance of
        # unpickleable things inside and possibly reduces
        # message size sent over the pipe
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            dummy = self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def close(self):
        self._handler.close()
        logging.Handler.close(self)

    def __del__(self):
        self.close()


class MultiProcRotatingFileHandler(MultiProcHandler):
    subhandler_cls = RotatingFileHandler

    def __init__(self, filename, mode="a", maxBytes=2 ** 20, backupCount=0):
        super().__init__(filename, mode, maxBytes, backupCount)


class MultiProcStreamHandler(MultiProcHandler):
    subhandler_cls = StreamHandler

    def __init__(self, stream):
        super().__init__(stream)


class MultiProcBufferingSMTPHandler(MultiProcHandler):
    subhandler_cls = BufferingSMTPHandler

    def __init__(
        self,
        capacity,
        mailhost,
        toaddrs,
        subject=None,
        flushLevel=ERROR,
        *,
        credentials=None,
        fromaddr=None,
        secure=None,
        mailport=None,
        timeout=5.0
    ):
        super().__init__(
            capacity=capacity,
            mailhost=mailhost,
            toaddrs=toaddrs,
            subject=subject,
            flushLevel=flushLevel,
            credentials=credentials,
            fromaddr=fromaddr,
            secure=secure,
            mailport=mailport,
            timeout=timeout,
        )



# coding:utf-8
import os
import ast
import ujson as json
from bourbaki.introspection.prettyprint import fmt_pyobj
from .exceptions import ConfigNotSerializable

DEFAULT_PY_INDENT = "    "
MAX_PY_WIDTH = 80


class ErroneousPythonSourceConfig:
    pass


class PythonSourceConfigSyntaxError(ErroneousPythonSourceConfig, SyntaxError):
    pass


class PythonSourceConfigRuntimeError(ErroneousPythonSourceConfig, RuntimeError):
    pass


class UnsafePythonSourceConfig(SyntaxError):
    pass


class IllegalFunctionInSourceConfig(UnsafePythonSourceConfig):
    pass


class IllegalExpressionInSourceConfig(UnsafePythonSourceConfig):
    pass


class IllegalNameInSourceConfig(UnsafePythonSourceConfig):
    pass


class PythonConfigNotSerializable(ConfigNotSerializable):
    pass


def is_json_serializable(conf):
    try:
        with open(os.devnull, "w") as f:
            json.dump(conf, f)
    except TypeError:
        return False
    else:
        return True


# safe python syntactic constructs for configuration
legal_python_config_exprs = (
    ast.Constant,
    ast.BinOp,
    ast.operator,
    ast.Compare,
    ast.Name,
    ast.Attribute,
    ast.List,
    ast.Tuple,
    ast.Dict,
    ast.Set,
    ast.Subscript,
    ast.ListComp,
    ast.DictComp,
    ast.SetComp,
    ast.comprehension,
    ast.keyword,
    ast.Assign,
    ast.Call,
)

# safe python builtin functions for configuration
legal_builtin_functions = {
    "abs",
    "all",
    "any",
    "ascii",
    "bin",
    "bool",
    "bytearray",
    "bytes",
    "callable",
    "chr",
    "complex",
    "dict",
    "dir",
    "divmod",
    "enumerate",
    "filter",
    "frozenset",
    "hash",
    "hex",
    "id",
    "int",
    "iter",
    "len",
    "list",
    "map",
    "max",
    "min",
    "next",
    "oct",
    "ord",
    "pow",
    "range",
    "repr",
    "reversed",
    "round",
    "set",
    "slice",
    "sorted",
    "str",
    "sum",
    "tuple",
    "vars",
    "zip",
}


illegal_builtin_names = {
    "__builtin__",
    "__builtins__",
    "__doc__",
    "__loader__",
    "__import__",
    "__name__",
    "__package__",
    "__spec__",
    "__file__",
    "__module__",
}


def validate_python_config_source(ast_or_source):
    if isinstance(ast_or_source, str):
        tree = ast.parse(ast_or_source)
    else:
        tree = ast_or_source

    msg = (
        "when using python source as configuration, supply only a single expression "
        "(generally a builtin collection expression) or a sequence of variable assignments; got {}"
    )
    mode = "exec"
    if len(tree.body) > 1:
        if not all(isinstance(node, ast.Assign) for node in tree.body):
            raise PythonSourceConfigSyntaxError(msg.format(list(map(type, tree.body))))
    elif len(tree.body) == 1:
        if not isinstance(tree.body[0], (ast.Expr, ast.Assign)):
            raise PythonSourceConfigSyntaxError(msg.format(type(tree.body[0])))
        if not isinstance(tree.body[0], ast.Assign):
            mode = "eval"

    def inner(node):
        if isinstance(node, (list, ast.Module)):
            if not isinstance(node, list):
                node = node.body
            return all(inner(v) for v in node)

        if isinstance(node, ast.Call):
            f = node.func
            msg = "function {} is not allowed in python configuration source; occurred at line {}, column {}"
            if isinstance(f, ast.Name) and f.id not in legal_builtin_functions:
                raise IllegalFunctionInSourceConfig(
                    msg.format(repr(node.func.id), node.lineno, node.col_offset)
                )
            elif not isinstance(f, ast.Name):
                raise IllegalFunctionInSourceConfig(
                    msg.format(repr(node.func), node.lineno, node.col_offset)
                )

        if isinstance(node, ast.Name):
            if node.id in illegal_builtin_names:
                raise IllegalNameInSourceConfig(
                    "reference to builtin name {} is not allowed in python configuration "
                    "source; occurred at line {}, column {}".format(
                        repr(node.id), node.lineno, node.col_offset
                    )
                )

        if isinstance(node, ast.Expr):
            return inner(node.value)

        if not isinstance(node, legal_python_config_exprs):
            raise IllegalExpressionInSourceConfig(
                "expression type {} is not allowed in python configuration source; "
                "occurred at line {}, column {}".format(
                    type(node), node.lineno, node.col_offset
                )
            )

        subexprs = tuple(getattr(node, attr) for attr in node._fields)
        subexprs = (
            e
            for e in subexprs
            if isinstance(e, (ast.AST, list))
            and not isinstance(e, (ast.Store, ast.Load))
        )

        return all(map(inner, subexprs))

    if inner(tree):
        if mode == "eval":
            tree = ast.Expression(tree.body[0].value)
        return tree, mode


def load_python(file):
    source = file.read()
    path = getattr(file, "name", "<string>")
    try:
        tree = ast.parse(source)
    except SyntaxError as e:
        raise PythonSourceConfigSyntaxError(
            "{}: {}; occurred on line {}, column {}".format(
                e.msg, repr(e.text), e.lineno, e.offset
            )
        )

    tree, mode = validate_python_config_source(tree)

    code = compile(tree, filename=path, mode=mode)

    def exec_(code, mode, locals):
        execmode = mode == "exec"
        eval_ = exec if execmode else eval
        try:
            conf = eval_(code, locals)
        except Exception as e:
            tb = e.__traceback__
            frame = tb.tb_frame
            while frame is not None and frame.f_code.co_filename != path:
                frame = frame.f_back
            raise PythonSourceConfigRuntimeError(
                "Couldn't evaluate python configuation source in {} mode: {}{}".format(
                    repr(mode),
                    e,
                    ""
                    if frame is None
                    else "; occurred at line {}".format(frame.f_lineno),
                ),
                tb,
            )
        return locals if execmode else conf

    conf = exec_(code, mode, {})
    conf.pop("__builtins__", None)
    return conf


def dump_python(conf, file, check_json_compat=False):
    if check_json_compat and not is_json_serializable(conf):
        raise PythonConfigNotSerializable(
            "can only write json-serializable config to python source; got {}".format(
                conf
            )
        )

    print(fmt_pyobj(conf, top_level=True), file=file)



# coding:utf-8
"""
Configure bash completions for argparse-based command line interfaces
Note: the full functioning of this module requires installation of [bash-completion](https://github.com/scop/bash-completion)
In turn, `bash-completion` assumes that GNU versions of common command line utilities are installed, so if you are using
a Mac OS, you will need to install these as well, using homebrew (`bash-completion` itself is also available via homebrew)
The minimal set set of dependencies can be installed via:
$ brew install bash-completion
$ brew install gnu-sed --with-default-names
$ brew install gnu-which --with-default-names
$ brew install grep --with-default-names
$ brew install findutils --with-default-names
$ brew install coreutils
$ brew install awk

You will also need to add the following line to your ~/.bash_profile:
$ [ -f /usr/local/etc/bash_completion ] && . /usr/local/etc/bash_completion
See the linked repository for `bash-completion` more detailed installation instructions for all operating systems.

>>> from argparse import ArgumentParser
>>> parser = ArgumentParser()
>>> # define parser arguments
>>> ...
>>> from bourbaki.application.completion import install_shell_completion
>>> install_shell_completion(parser, "command_for_script", "other_command_for_script", ...)

Or, if your parser is an `bourbaki.application.CommandLineInterface`, just
>>> parser.install_shell_completion()
In that case, you may optionally supply command names; otherwise they are inferred from the `prog` arg to
the constructor as in `argparse.ArgumentParser` (an error is thrown if no `prog` is passed)
"""

from .compgen_python_classpaths import COMPLETION_DEBUG_ENV_VAR
from .completers import install_shell_completion
from .completers import (
    Complete,
    RawShellFunctionComplete,
    FixedShellFunctionComplete,
    BashCompletion,
)



# coding:utf-8
from typing import Union, Dict, Mapping, Any, Optional
import os
import io
import sys
from pathlib import Path
from warnings import warn
from logging import Logger, StreamHandler
from _io import TextIOWrapper
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
from logging.config import dictConfig
from datetime import datetime
from .helpers import validate_log_level_int, validate_log_level_str
from .regexes import log_fields
from .defaults import (
    DEFAULT_FILE_HANDLER_CONFIG,
    DEFAULT_CONSOLE_HANDLER_CONFIG,
    DEFAULT_SMTP_HANDLER_CONFIG,
)
from .defaults import (
    DEFAULT_FILE_HANDLER_MP_CONFIG,
    DEFAULT_CONSOLE_HANDLER_MP_CONFIG,
    DEFAULT_SMTP_HANDLER_MP_CONFIG,
)
from .defaults import (
    DEFAULT_LOG_DATE_FMT,
    DEFAULT_LOG_MSG_FMT,
    DEFAULT_LOG_MSG_MP_FMT,
    VERBOSE_LOG_MSG_FMT,
)
from bourbaki.application.config import load_config
from bourbaki.application.paths import DEFAULT_FILENAME_DATE_FMT
from .defaults import (
    DEFAULT_FILE_LOG_LEVEL,
    DEFAULT_CONSOLE_LOG_LEVEL,
    DEFAULT_SMTP_LOG_LEVEL,
    empty_config_dict,
)


__doc__ = """
    Application-level logging configuration functions
    example, to configure logging globally, run the following at the start of a script:
    
    logfile = "/path/to/my/log_dir/MyApp.log"
    
    logconfig = default_config(logfile, console=TRUE)  # here we specify that we log to a file and the console
    configure_logging(logconfig, dated_logfiles=False)  # here we specify that log files will not have run times
                                                        # appended to their names
    
    # now any class that gets a logger as a child of the root logger
    # will have a logger attached that logs to the correct places, e.g.:
    
    class MyClass(Logged):
        __instance_naming__ = 'int'
        __verbose__ = True  # log constructor errors
        
        def __init__(self, *args):
            # a DEBUG-level statement is logged prior to entering this function thanks to the __verbose__ flag
            # and an ERROR-level statement will be logged if it fails
            self.args = args
    
    my_instance = MyClass(1, 2, 3)
    my_instance.logger.info("hello!")
    
    will print something like this to the console and/or log file:
    
    MyClass-1: <line no> <datetime> DEBUG - initializing new instance
    MyClass-1: <line no> <datetime> INFO - hello!
    """

LogLevel = Union[str, int]
IOSpec = Union[str, TextIOWrapper]

instance_naming_opts = ("int", "hex", "keyword", None)


def make_log_format(fields, sep=" - "):
    """
    Build a custom log format, as accepted by the logging module, from a list of field names.
    :param fields: list or tuple of str - names of fields to use in log messages
    :param sep: str - separator to put between fields. Default is ' - '
    :return: a log format string usable to configure log formatters
    """
    assert all(f in log_fields for f in fields), "Only fields from {} are valid".format(
        tuple(log_fields)
    )
    return sep.join("%({}){}".format(f, log_fields[f]) for f in fields)


def configure_default_logging(
    filename=None,
    console=True,
    verbose_format=False,
    smtp_credentials=None,
    smtp_to_addrs=None,
    smtp_mailhost=None,
    smtp_tls_credentials=(),
    smtp_buffer_capacity=10,
    file_level=DEFAULT_FILE_LOG_LEVEL,
    console_level=DEFAULT_CONSOLE_LOG_LEVEL,
    smtp_level=DEFAULT_SMTP_LOG_LEVEL,
    multiprocessing=False,
    dated_logfiles=False,
    disable_existing_loggers=False,
):
    """
    Batteries-included default configuration in one line.

    :param filename: Optional path to a file to log to, if file logging is desired
    :param console: boolean indicating whether to log to the console
    :param level: string ("DEBUG". "INFO", "WARN", "ERROR", "CRITICAL") or int (0-50)
        (or a custom level registered in an imported package, i.e. logging defines "PROGRESS" > "INFO")
        The lowest level of any log message that will be printed to file or console.
    :param dated: bool - append datetime to log filename? Default is False.
    :return: None; this configures logging globally in the logging module
    """
    if verbose_format:
        fmt = VERBOSE_LOG_MSG_FMT
    else:
        fmt = DEFAULT_LOG_MSG_FMT if not multiprocessing else DEFAULT_LOG_MSG_MP_FMT

    conf = default_config(
        filename,
        console,
        format=fmt,
        smtp_credentials=smtp_credentials,
        smtp_to_addrs=smtp_to_addrs,
        smtp_mailhost=smtp_mailhost,
        smtp_tls_credentials=smtp_tls_credentials,
        smtp_buffer_capacity=smtp_buffer_capacity,
        file_log_level=file_level,
        console_log_level=console_level,
        smtp_log_level=smtp_level,
        multiprocessing=multiprocessing,
        disable_existing_loggers=disable_existing_loggers,
        dated_logfiles=dated_logfiles,
    )

    configure_custom(conf)


def configure_debug_logging(
    filename=None, dated_logfiles=True, disable_existing_loggers=False
):
    """
    Batteries-included debug-level logging configuration in one line.
    All loggers are set to 'DEBUG' log level and log to console as well as a file, if
    a filename is specified.

    :param filename: Optional path to a file to log to, if file logging is desired
    :param dated: bool - append datetime to log filename? Default is True, to create a
      new timestamped file for a fresh run.
    :return: None; this configures logging globally in the logging module
    """
    conf = default_config(
        filename=filename,
        console=True,
        file_log_level=DEBUG,
        console_log_level=DEBUG,
        dated_logfiles=dated_logfiles,
        disable_existing_loggers=disable_existing_loggers,
    )

    configure_custom(conf)


def configure_custom(
    config: Union[str, Dict[str, Any]], disable_existing_loggers: Optional[bool] = None
):
    """
    Configure the global logging properties for a run of an application

    :param config: dictionary as passed to logging.config.dictConfig
        OR a file path, which case the config dict is read from a file (.yml or .json)
    :param disable_existing_loggers: boolean. When logging.config.dictConfig is called, it disables
        all existing loggers by default. When this is False, they are re-enabled after the configuration
        to allow pre-configured loggers to continue logging to their specified locations
    :return: None; this configures logging globally in the logging module
    """
    if isinstance(config, (str, Path, io.IOBase)):
        config = load_config_from_file(config)

    elif not isinstance(config, Mapping):
        raise ValueError(
            "config must be a path to a config file or a dictionary as accepted by "
            "logging.config.dictConfig"
        )

    if (
        disable_existing_loggers is not None
        and config.get("disable_existing_loggers") != disable_existing_loggers
    ):
        config = config.copy()
        config["disable_existing_loggers"] = disable_existing_loggers

    dictConfig(config)


configure_custom_logging = configure_custom


def load_config_from_file(filename) -> Dict[str, Any]:
    conf = load_config(filename)

    if not isinstance(conf, dict):
        warn(
            "Warning: the loaded config does not appear to be a key-value mapping; a call to "
            "logging.config.dictConfig or configure_logging will fail."
        )

    return conf


def default_config(
    filename: Optional[str] = None,
    console: bool = True,
    file_log_level: LogLevel = DEFAULT_FILE_LOG_LEVEL,
    console_log_level: LogLevel = DEFAULT_CONSOLE_LOG_LEVEL,
    smtp_log_level: LogLevel = DEFAULT_SMTP_LOG_LEVEL,
    root_log_level: Optional[LogLevel] = None,
    smtp_credentials=None,
    smtp_to_addrs=None,
    smtp_mailhost=None,
    smtp_tls_credentials=(),
    smtp_buffer_capacity=10,
    format: str = DEFAULT_LOG_MSG_FMT,
    datefmt: str = DEFAULT_LOG_DATE_FMT,
    stream: IOSpec = "stderr",
    multiprocessing=False,
    dated_logfiles: bool = False,
    max_logfile_bytes: int = 2 ** 30,
    disable_existing_loggers: bool = True,
    check_dir: bool = True,
):
    """
    Get a dictionary for configuring logging with a file logger and a console logger.
    The console logger is by default configured at the DEBUG level, while
    the file logger only logs messages at the INFO level and above.
    All args documented here have reasonable defaults, so you need not worry about most of them.

    :param filename: where to log to on disk. If None, don't configure a file logger
    :param console: bool indicating whether to log to the terminal stdout. Default is True
    :param file_log_level: int or string indicating the logging level of the file logger (if filename is given).
        Default is "INFO"
    :param root_log_level: int or string indicating the logging level of the root logger. This acts as a filter for
        all other loggers, so use wisely. Default is "DEBUG", meaning that all messages will come through
    :param stream: either a str in ('stdout', 'stdin') or io.TextIOWrapper such as sys.stdout,
    :param format: format of logger messages, as passed to the logging.Formatter constructor
    :param datefmt: format of dates in logger messages, as passed to the logging.Formatter constructor
    :param max_logfile_bytes: max size of the log file in bytes. Default is 2**30 = 1Gb
    :param check_dir: bool indicating whether to check if the directory of the logfile exists. An exception is
        raised if not
    :return: a dict suitable for passing to logging.config.dictConfig
    """
    file_log_level = validate_log_level_int(file_log_level or NOTSET)
    smtp_log_level = validate_log_level_int(smtp_log_level or NOTSET)
    console_log_level = validate_log_level_int(console_log_level or NOTSET)
    if root_log_level is None:
        root_log_level = min([console_log_level, file_log_level, smtp_log_level])
    else:
        root_log_level = validate_log_level_int(root_log_level or NOTSET)

    # construct a copy of an empty config each time to avoid mutation problems
    config = empty_config_dict.copy()
    config["formatters"]["default"] = dict(format=format, datefmt=datefmt)
    config["root"]["level"] = root_log_level

    if console:
        console_config = (
            DEFAULT_CONSOLE_HANDLER_CONFIG.copy()
            if not multiprocessing
            else DEFAULT_CONSOLE_HANDLER_MP_CONFIG.copy()
        )
        if stream is not None:
            console_config["stream"] = get_stream_name_for_config(stream)
        console_config["level"] = console_log_level
        config["handlers"]["console"] = console_config

    if filename is not None:
        assert isinstance(filename, str), "filename must be a string, not {}".format(
            type(filename)
        )
        filename = os.path.abspath(filename)

        if check_dir:
            logdir = os.path.dirname(filename)
            assert os.path.isdir(logdir), "the directory '{}' does not exist".format(
                logdir
            )
        if max_logfile_bytes is not None:
            assert isinstance(max_logfile_bytes, int), "max_bytes must be an int"

        if dated_logfiles:
            filename = append_log_date(filename)

        fileconfig = (
            DEFAULT_FILE_HANDLER_CONFIG.copy()
            if not multiprocessing
            else DEFAULT_FILE_HANDLER_MP_CONFIG.copy()
        )

        fileconfig["level"] = file_log_level
        fileconfig["filename"] = filename
        fileconfig["maxBytes"] = max_logfile_bytes
        config["handlers"]["file"] = fileconfig

    if smtp_mailhost is not None or smtp_credentials is not None:
        smtp_config = (
            DEFAULT_SMTP_HANDLER_CONFIG.copy()
            if not multiprocessing
            else DEFAULT_SMTP_HANDLER_MP_CONFIG.copy()
        )

        if smtp_to_addrs is None:
            if isinstance(smtp_credentials, (list, tuple)):
                smtp_to_addrs = [smtp_credentials[0]]
            elif isinstance(smtp_credentials, str):
                smtp_to_addrs = [smtp_credentials]

        for name, var in [
            ("level", smtp_log_level),
            ("capacity", smtp_buffer_capacity),
            ("mailhost", smtp_mailhost),
            ("credentials", smtp_credentials),
            ("toaddrs", smtp_to_addrs),
            ("secure", smtp_tls_credentials),
        ]:
            if var is not None:
                smtp_config[name] = var

        config["handlers"]["email"] = smtp_config

    # root logger uses all handlers
    config["root"]["handlers"] = list(config["handlers"].keys())

    config["disable_existing_loggers"] = disable_existing_loggers

    return config


def enable_loggers(logger_cls=Logger):
    for logger in logger_cls.manager.loggerDict.values():
        logger.disabled = False


def disable_loggers(logger_cls=Logger):
    for logger in logger_cls.manager.loggerDict.values():
        logger.disabled = True


def all_loggers():
    def inner():
        yield Logger.root
        for log in Logger.manager.loggerDict.values():
            yield log

    return list(inner())


def logs_to_console(logger):
    def _logs_to_console(handler):
        return handler.stream.name in ("<stderr>", "<stdout>", "stderr", "stdout")

    if isinstance(logger, StreamHandler):
        return _logs_to_console(logger)
    elif isinstance(logger, Logger):
        return any(_logs_to_console(h) for h in logger.handlers)


def disable_console_logging(*loggers):
    loggers = loggers or all_loggers()
    for logger in loggers:
        if not hasattr(logger, "handlers"):
            continue
        cache = []
        handlers = []
        for handler in logger.handlers:
            if logs_to_console(handler):
                cache.append(handler)
            else:
                handlers.append(handler)
        if cache:
            logger.handlers = handlers
            logger._handler_cache = cache


def enable_console_logging(*loggers):
    loggers = loggers or all_loggers()
    for logger in loggers:
        cache = getattr(logger, "_handler_cache", None)
        if cache:
            logger.handlers.extend(cache)
            delattr(logger, "_handler_cache")


def get_stream_name_for_config(stream: IOSpec) -> Union[str, TextIOWrapper]:
    if isinstance(stream, str):
        assert stream in (
            "stdout",
            "stderr",
        ), "if stream is a string, it must be in ('stdout', 'stderr')"
    else:
        assert isinstance(
            stream, TextIOWrapper
        ), "if stream is not a string, it must be an instance of _io.TextIOWrapper"

    if stream in ("stdout", sys.stdout):
        return "ext://sys.stdout"
    elif stream in ("stderr", sys.stderr):
        return "ext://sys.stderr"
    else:
        return stream


def append_log_date(filename: str, datestr=None, sep="") -> str:
    cur_time_str = datestr or datetime.now().strftime(DEFAULT_FILENAME_DATE_FMT)
    name, ext = os.path.splitext(filename)
    return name + sep + cur_time_str + ext



# coding:utf-8
from typing import IO, List, Tuple, Sequence, Union, Optional as Opt
from abc import ABC
import argparse
from argparse import ArgumentParser, Action, FileType, _SubParsersAction
from functools import singledispatch, lru_cache
import os
from pathlib import Path
import pkgutil
import re
from shlex import quote
import sys
from warnings import warn

from bourbaki.introspection.classes import parameterized_classpath
from bourbaki.introspection.generic_dispatch import const

from ..paths import is_newer
from .compgen_python_classpaths import (
    MODULE_FLAG,
    CALLABLE_FLAG,
    CLASS_FLAG,
    INSTANCE_FLAG,
    SUBCLASS_FLAG,
)

REPEATABLE_ACTIONS = tuple(
    cls
    for cls in vars(argparse).values()
    if isinstance(cls, type)
    and issubclass(cls, Action)
    and ("Count" in cls.__name__ or "Append" in cls.__name__)
)

BASH_COMPLETION_HELPERS_FILENAME = "bourbaki_bash_completion_helpers.sh"
BASH_COMPLETION_FUNC_NAME = "_bourbaki_complete"
COMPLETE_CHOICES_FUNC_NAME = "_bourbaki_complete_choices"
COMPLETE_FILES_FUNC_NAME = "_bourbaki_complete_files"
COMPLETE_CLASSPATHS_FUNC_NAME = "_bourbaki_complete_python_classpaths"
COMPLETE_INTS_FUNC_NAME = "_bourbaki_complete_ints"
COMPLETE_BOOLS_FUNC_NAME = "_bourbaki_complete_bools"
COMPLETE_FLOATS_FUNC_NAME = "_bourbaki_complete_floats"
COMPLETE_KEYVAL_FUNC_NAME = "_bourbaki_complete_keyval"
NO_COMPLETE_FUNC_NAME = "_bourbaki_no_complete"
KEYVAL_SEP = "="
OPTION_SEP = ","
BASH_COMPLETION_TREE_INDENT = "  "

DEFAULT_BASH_COMPLETE_OPTIONS = ("bashdefault", "filenames")
BASH_SHEBANG = "#!/usr/bin/env bash"
BASH_SOURCE_TEMPLATE = "[ -f {} ] && source {}"

BASH_COMPLETION_USER_FILENAME = ".bash_completion"
BASH_COMPLETION_USER_DIRNAME = ".bash_completion.d"
BASH_COMPLETION_FUNCTIONS = dict(
    _filedir="files and directories",
    _signals="signal names",
    _mac_addresses="known mac addresses",
    _configured_interfaces="configured network interfaces",
    _kernel_versions="available kernels",
    _available_interfaces="all available network interfaces",
    _pids="process IDs",
    _pgids="process group IDs",
    _pnames="process names",
    _uids="user IDs",
    _gids="group IDs",
    _usergroup="user or user:group format",
    _services="services",
    _modules="modules",
    _installed_modules="installed modules",
    _shells="valid shells",
    _fstypes="valid filesystem types",
    _pci_ids="PCI IDs",
    _usb_ids="USB IDs",
    _cd_devices="CD device names",
    _dvd_devices="DVD device names",
    _function="shell functions",
    _user_at_host="user@host",
    _known_hosts_real="hosts based on ssh's config and known_hosts",
)


def shellquote(s: str):
    return quote(s)


def install_shell_completion(
    parser: ArgumentParser,
    *commands: str,
    extra_completion_script: Opt[str] = None,
    completion_options: Sequence[str] = DEFAULT_BASH_COMPLETE_OPTIONS,
    last_edit_time: Opt[Union[float, str, Path]] = None
):
    completions_file, completions_dir, helpers_file = (
        install_application_shell_completion()
    )
    cmdname = _shortest_identifier(*commands)
    custom_file = os.path.join(completions_dir, cmdname + ".sh")
    custom_file_abs = os.path.expanduser(custom_file)
    completions_file_abs = os.path.expanduser(completions_file)

    if last_edit_time is None or is_newer(last_edit_time, custom_file_abs):
        with open(custom_file_abs, "w") as outfile:
            write_bash_completion_for_parser(
                parser,
                outfile,
                commands=commands,
                completion_options=completion_options,
                extra_completion_script=extra_completion_script,
            )

    custom_source_command = BASH_SOURCE_TEMPLATE.format(custom_file, custom_file)
    _ensure_lines([custom_source_command], completions_file_abs)
    print("eval the following in your shell to update completions:", file=sys.stderr)
    print(custom_source_command, file=sys.stdout)


def install_application_shell_completion():
    # install the custom function in custom_file in this dir,
    completions_dir = get_user_bash_completion_dir()
    # then source it in here
    completions_file = get_user_bash_completion_path()
    # and also make sure the helpers are up-to-date here
    helpers_file = get_application_bash_completion_helpers_path()
    # only copies if the source version is modified more recently than the extant version
    _install_shell_completion_helpers(helpers_file)
    # ensure the source command is present
    helpers_source_command = BASH_SOURCE_TEMPLATE.format(helpers_file, helpers_file)
    _ensure_lines([helpers_source_command], completions_file)
    return completions_file, completions_dir, helpers_file


class Complete(ABC):
    args = None
    _shell_func_name = None

    def __str__(self):
        if not self.args:
            return self._shell_func_name or ""

        args = map(shellquote, map(str, self.args))
        try:
            if not self._shell_func_name:
                return " ".join(args)
            else:
                return "{} {}".format(self._shell_func_name, " ".join(args))
        except Exception as e:
            warn("Couldn't represent {} as a shell command: {}".format(repr(self), e))

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ", ".join(map(repr, self.args)))

    def to_bash_call(self):
        return str(self)


class RawShellFunctionComplete(Complete):
    def __init__(self, shell_function: str, *args: str):
        self._shell_func_name = shell_function
        self.args = args


class FixedShellFunctionComplete(Complete):
    def __init__(self, *args: str):
        self.args = args


class CompleteFromStdout(FixedShellFunctionComplete):
    """Complete from a shell command that prints lines to stdout"""
    _shell_func_name = '_bourbaki_complete_from_stdout'


class _BashCompletionCompleters:
    """Constuct completers from the `bash_completion` bash module, which are highly developed and heavily tested.
    This uses simple .attribute access using '_'-stripped bash_completion shell function names."""

    def __dir__(self):
        return [name.lstrip("_") for name in BASH_COMPLETION_FUNCTIONS]

    @lru_cache(None)
    def __getattr__(self, item: str):
        func_name = "_" + item
        if func_name not in BASH_COMPLETION_FUNCTIONS:
            warn(
                "'{}' is not registered as a know bash completion function".format(
                    func_name
                )
            )

        # always const, so that cli_completer.register(some_type)(BashCompletion.some_completer_name) just works
        bash_completion_cls = const(RawShellFunctionComplete(func_name))
        return bash_completion_cls


BashCompletion = _BashCompletionCompleters()


class CompleteFiles(Complete):
    _shell_func_name = COMPLETE_FILES_FUNC_NAME

    def __init__(self, *exts: str):
        self.args = tuple(e.lstrip(".") for e in exts)


class CompleteDirs(CompleteFiles):
    def __init__(self):
        # from bash completion docs: "If `-d', complete only on directories"
        super().__init__("-d")


class CompleteFilesAndDirs(CompleteFiles):
    def __init__(self, *exts: str):
        super().__init__("-d", *exts)


class CompletePythonClassPaths(FixedShellFunctionComplete):
    _shell_func_name = COMPLETE_CLASSPATHS_FUNC_NAME

    def __init__(self, *module_prefixes: str):
        super().__init__(*module_prefixes)


class _CompletePythonPathsWithPrefixes(CompletePythonClassPaths):
    _flag = None

    def __init__(self, *prefixes):
        super().__init__(self._flag, *prefixes)


class _CompletePythonPathsWithTypes(CompletePythonClassPaths):
    _flag = None

    def __init__(self, *superclasses):
        super().__init__(
            self._flag,
            *(parameterized_classpath(cls).replace(" ", "") for cls in superclasses)
        )


class CompletePythonClasses(_CompletePythonPathsWithPrefixes):
    _flag = CLASS_FLAG


class CompletePythonModules(_CompletePythonPathsWithPrefixes):
    _flag = MODULE_FLAG


class CompletePythonCallables(_CompletePythonPathsWithPrefixes):
    _flag = CALLABLE_FLAG


class CompletePythonInstances(_CompletePythonPathsWithTypes):
    _flag = INSTANCE_FLAG


class CompletePythonSubclasses(_CompletePythonPathsWithTypes):
    _flag = SUBCLASS_FLAG


class CompleteChoices(FixedShellFunctionComplete):
    _shell_func_name = COMPLETE_CHOICES_FUNC_NAME

    def __init__(self, *choices: str):
        super().__init__(*choices)


class CompleteEnum(CompleteChoices):
    def __init__(self, enum):
        super().__init__(*(e.name for e in enum))


class CompleteInts(FixedShellFunctionComplete):
    _shell_func_name = COMPLETE_INTS_FUNC_NAME


class CompleteBools(FixedShellFunctionComplete):
    _shell_func_name = COMPLETE_BOOLS_FUNC_NAME


class CompleteFloats(FixedShellFunctionComplete):
    _shell_func_name = COMPLETE_FLOATS_FUNC_NAME


class _NoComplete(FixedShellFunctionComplete):
    _shell_func_name = NO_COMPLETE_FUNC_NAME


NoComplete = _NoComplete()


class _MultiComplete(Complete):
    sep = None
    unique = False

    def __init__(self, *completers: Complete):
        self.completers = completers

    def __str__(self):
        collect = set if self.unique else iter
        return self.sep.join(collect(map(str, self.completers)))

    def __repr__(self):
        return "{}{}".format(self.__class__.__name__, repr(tuple(self.completers)))


class CompleteKeyValues(_MultiComplete):
    sep = " {} ".format(KEYVAL_SEP * 3)
    _shell_func_name = COMPLETE_KEYVAL_FUNC_NAME

    def __init__(self, key_completer, val_completer, *args):
        if args:
            warn(
                "got {} extra args for CompleteKeyValues; {} will be ignored".format(
                    len(args), repr(args)
                )
            )
        if key_completer is None:
            key_completer = NoComplete
        if val_completer is None:
            val_completer = NoComplete
        super().__init__(key_completer, val_completer)

    def __str__(self):
        return " ".join([self._shell_func_name, super().__str__()])


class CompleteTuple(_MultiComplete):
    sep = "; "


class CompleteUnion(_MultiComplete):
    sep = " && "
    unique = True


def write_bash_completion_for_parser(
    parser: ArgumentParser,
    file: IO[str],
    commands: Sequence[str],
    completion_options: Opt[Union[str, Sequence[str]]] = None,
    extra_completion_script: Opt[str] = None,
    shebang: Opt[str] = BASH_SHEBANG,
):
    if isinstance(commands, str):
        commands = [commands]

    if completion_options:
        optionstr = (
            " -o ".join(completion_options)
            if not isinstance(completion_options, str)
            else completion_options
        )
        optionstr = "-o {} ".format(optionstr)
    else:
        optionstr = ""

    def print_(*line):
        print(*line, file=file)
        print(*line, file=sys.stderr)

    print("WRITING LINES TO {}:".format(file.name), file=sys.stderr)
    if shebang:
        print_(shebang)
    print_("")

    name = _shortest_identifier(*commands)
    completion_funcname = "_complete_{}".format(name)

    print_("{}() {{".format(completion_funcname))
    print_('{} """'.format(BASH_COMPLETION_FUNC_NAME))
    print_cli_def_tree(parser, file=file)
    print_('"""')
    print_("}\n")

    for cmd in commands:
        line = "complete {}-F {} {}".format(optionstr, completion_funcname, cmd)
        print_(line)

    if extra_completion_script:
        print_()
        print_(extra_completion_script)

    print_()


def print_cli_def_tree(
    parser: Union[ArgumentParser, _SubParsersAction],
    file: IO[str] = sys.stdout,
    indent: str = "",
):
    args, options, commands = gather_args_options_subparsers(parser)

    def print_(*line):
        print(*line, file=file)
        print(*line, file=sys.stderr)

    for a in args:
        print_("{}- {}".format(indent, completion_spec(a, positional=True)))

    for a in options:
        print_(
            "{}{} {}".format(
                indent,
                OPTION_SEP.join(a.option_strings),
                completion_spec(a, positional=False),
            )
        )

    for name, c in commands:
        print_("{}{}".format(indent, name))
        print_cli_def_tree(c, file, indent=indent + BASH_COMPLETION_TREE_INDENT)


def gather_args_options_subparsers(
    parser: ArgumentParser
) -> Tuple[List[Action], List[Action], List[Tuple[str, Action]]]:
    args = []
    options = []
    commands = []
    for action in parser._actions:
        if action.option_strings:
            options.append(action)
        elif isinstance(action, _SubParsersAction):
            if action.choices:
                for tup in action.choices.items():
                    commands.append(tup)
        else:
            args.append(action)

    return args, options, commands


def completion_spec(action: Action, positional=False) -> str:
    if isinstance(action, REPEATABLE_ACTIONS):
        nreps = "+" if action.required else "*"
    elif positional:
        nreps = None
    else:
        nreps = "1" if action.required else "?"

    nargs = action.nargs if action.nargs is not None else 1
    completer = get_completer(action) if nargs != 0 else None
    repstr = "{}{}".format(nargs, "" if nreps is None else "({})".format(nreps))
    return (
        "{} {}".format(repstr, bash_call_str(completer))
        if completer is not None
        else repstr
    )


def get_completer(action: Action) -> Opt[Union[str, List[str], Complete]]:
    if action.choices:
        completer = CompleteChoices(*action.choices)
    else:
        completer = getattr(action, "completer", None)
        if completer is None and action.nargs != 0:
            completer = completer_argparser_from_type(action.type)

    return completer


@singledispatch
def completer_argparser_from_type(t):
    # this handles any user-monkey-patched types with 'completer/_completer' attribute
    return getattr(t, "_completer", getattr(t, "completer", None))


@completer_argparser_from_type.register(FileType)
def completer_from_type_argparse_file(t):
    return CompleteFiles()


@singledispatch
def bash_call_str(completer: Union[str, Complete]):
    return str(completer)


@bash_call_str.register(list)
def bash_call_str_args_list(args):
    return " ".join(map(shellquote, args))


def get_user_bash_completion_path() -> str:
    name = str(Path("~") / BASH_COMPLETION_USER_FILENAME)
    absolute = os.path.expanduser(name)
    if os.path.exists(absolute) and not os.path.isfile(absolute):
        raise FileExistsError(
            "{} is not a file but is required to be by application bash completion"
        )
    elif not os.path.exists(absolute):
        # touch the file
        with open(absolute, "a"):
            pass
    return name


def get_user_bash_completion_dir() -> str:
    name = str(Path("~") / BASH_COMPLETION_USER_DIRNAME)
    absolute = os.path.expanduser(name)
    if not os.path.exists(absolute):
        os.mkdir(absolute)
    elif not os.path.isdir(absolute):
        raise NotADirectoryError(
            "{} is not a directory but is required to be by application bash completion"
        )

    return name


def get_application_bash_completion_helpers_path() -> str:
    completions_dir = get_user_bash_completion_dir()
    return str(Path(completions_dir) / BASH_COMPLETION_HELPERS_FILENAME)


def _shortest_identifier(*commands: str):
    def to_identifier(name):
        return re.sub(r"[^.\w]", "_", name)

    return min(map(to_identifier, commands), key=len)


def _ensure_lines(lines, textfile: str):
    textfile = os.path.expanduser(textfile)
    print("ENSURING LINES ARE PRESENT IN {}:".format(textfile), file=sys.stderr)
    print("\n".join(lines), file=sys.stderr)
    print(file=sys.stderr)
    missing = set(lines)
    present = set()

    if os.path.exists(textfile):
        with open(textfile, "r") as outfile:
            for line in map(str.strip, outfile):
                if line in missing:
                    present.add(line)
                    missing.remove(line)
        # ensure consistent order here
        to_write = [l for l in lines if l in missing]
    else:
        to_write = lines

    if to_write:
        print("WRITING LINES TO {}:".format(textfile), file=sys.stderr)
        with open(textfile, "a") as outfile:
            for line in to_write:
                print(line, file=outfile)
                print(line, file=sys.stderr)
        print(file=sys.stderr)


def _install_shell_completion_helpers(completions_helpers_file):
    source = pkgutil.get_data(
        "bourbaki.application.completion", BASH_COMPLETION_HELPERS_FILENAME
    ).strip()
    completions_helpers_file_absolute = os.path.expanduser(completions_helpers_file)

    if not os.path.exists(completions_helpers_file_absolute):
        with open(completions_helpers_file_absolute, "wb") as outfile:
            print(
                "INSTALLING BASH COMPLETION HELPERS AT {}".format(
                    completions_helpers_file_absolute
                ),
                file=sys.stderr,
            )
            outfile.write(source)
    else:
        with open(completions_helpers_file_absolute, "rb+") as outfile:
            old_source = outfile.read().strip()
            if source != old_source:
                print(
                    "REINSTALLING BASH COMPLETION HELPERS AT {}".format(
                        completions_helpers_file_absolute
                    ),
                    file=sys.stderr,
                )
                outfile.seek(0)
                outfile.truncate()
                outfile.write(source)



# coding:utf-8
import os
from itertools import chain
from pathlib import Path
from setuptools import find_packages
import pytest

module = type(os)

top_dir = Path(__file__).parent.parent


def all_module_paths(rootdir):
    parents = find_packages(rootdir)
    return list(chain(parents, chain.from_iterable(map(submodule_paths, parents))))


def submodule_paths(parent_module_path: str):
    paths = Path(parent_module_path.replace(".", os.path.sep)).glob("*.py")
    return (
        parent_module_path + "." + os.path.splitext(p.name)[0].replace(os.path.sep, ".")
        for p in paths
        if not p.name.startswith("_")
    )


@pytest.mark.parametrize("module_name", all_module_paths(top_dir))
def test_submodule_import(module_name):
    mod = __import__(module_name)
    assert isinstance(mod, module)



# coding:utf-8
from typing import Callable
from types import MethodType
from enum import Enum
from functools import wraps
from inspect import signature
from logging import Logger, addLevelName, getLoggerClass
from . import defaults

addLevelName(defaults.TRACE, defaults.TRACE_LEVEL)

LOG_NAME_KWARG = "__logname__"


class InstanceLoggerNamingConvention(Enum):
    int = "append an incremented id to the class name for instance logger names"
    hexaddress = "append the hexadecimal memory address to the class name for instance logger names"
    keyword = (
        "use the keyword arg %s to pass in a custom instance logger name via the constructor"
        % LOG_NAME_KWARG
    )


class Logged:
    __logger__ = None
    __log_level__ = None
    __verbose__ = False
    __log_tracebacks__ = True
    __instance_naming__ = None
    __last_instance_id__ = None

    @property
    def __logname__(self):
        name = self.__dict__.get("__logname__")
        if name is None:
            convention = self.__instance_naming__
            if convention is None:
                cls = type(self)
                name = cls.__qualname__
            elif isinstance(convention, InstanceLoggerNamingConvention):
                cls = type(self)
                clsname = cls.__qualname__
                if convention is InstanceLoggerNamingConvention.int:
                    lastid = cls.__last_instance_id__ + 1
                    name = "{}-{}".format(clsname, lastid)
                    __last_instance_id__ = lastid
                elif convention is InstanceLoggerNamingConvention.hexaddress:
                    id_ = hex(id(self))
                    name = "{}@{}".format(clsname, id_)
                else:
                    # keyword option, but not passed to constructor
                    name = self.__dict__.get("__logname__", clsname)
            elif callable(convention):
                if isinstance(convention, MethodType):
                    # bound method from function defined in the namespace of the class
                    name = convention()
                else:
                    # attribute set at __init__ as function literal, or some other callable set at the class level
                    name = convention(self)
            else:
                raise TypeError(
                    "__instance_naming__ attribute must be an instance of {} or a callable; got {}".format(
                        InstanceLoggerNamingConvention, type(convention)
                    )
                )

        return name

    @__logname__.setter
    def __logname__(self, name):
        if not isinstance(name, str):
            raise TypeError("__logname__ must be a string")
        self.__dict__["__logname__"] = name
        logger = self.__dict__.get("__logger__")
        if logger is not None:
            logger.name = name

    @property
    def __logger_cls__(self):
        return getLoggerClass()

    def __init_subclass__(cls, **kwargs):
        convention = cls.__instance_naming__
        if convention is None:
            pass
        elif isinstance(convention, str):
            convention = getattr(InstanceLoggerNamingConvention, convention, None)
            if convention is None:
                raise ValueError(
                    "__instance_naming__ attribute, if a string, must be one of {}".format(
                        [c.name for c in InstanceLoggerNamingConvention]
                    )
                )
        elif not callable(convention) and not isinstance(
            convention, InstanceLoggerNamingConvention
        ):
            raise TypeError(
                "__instance_naming__ attribute must be either a callable taking a single instance of "
                "{} or an {} instance".format(cls, InstanceLoggerNamingConvention)
            )

        if cls.__verbose__ or convention is InstanceLoggerNamingConvention.keyword:
            cls.__init__ = verbose_init(cls.__init__)

        if convention is InstanceLoggerNamingConvention.int:
            cls.__last_instance_id__ = 0

        if cls.__verbose__:
            cls.__signature__ = signature(cls)

        cls.__instance_naming__ = convention

    @property
    def logger(self):
        logger = self.__dict__.get("__logger__", None)
        if logger is None:
            # construct an instance logger
            instance_name = self.__logname__
            logger_cls = self.__logger_cls__
            logger = logger_cls.manager.getLogger(instance_name)
            if self.__log_level__ is not None:
                logger.setLevel(self.__log_level__)
            self.__dict__["__logger__"] = logger

        return logger

    @logger.setter
    def logger(obj, logger):
        if not isinstance(logger, Logger):
            raise TypeError(
                "class logger must be an instance of logging.Logger; got {}".format(
                    type(logger)
                )
            )
        obj.__dict__["__logger__"] = logger

    def __getstate__(self):
        self.logger.debug(
            "Getting state for pickling/copying, excluding __logger__ attribute"
        )
        # can't pickle RThreadLock objects!
        state = self.__dict__.copy()
        state.pop("__logger__", None)
        self.logger.debug("Got state dict with keys %s", state.keys())
        return state


def verbose_init(init: Callable):
    @wraps(init)
    def __init__(self: Logged, *args, **kwargs):
        if self.__instance_naming__ is InstanceLoggerNamingConvention.keyword:
            name = kwargs.pop(LOG_NAME_KWARG, None)
            if name is not None:
                self.__logname__ = name
        if self.__verbose__:
            logger = self.logger
            cls = type(self)
            logger.debug("Initializing %s instance", cls)
            try:
                init(self, *args, **kwargs)
            except Exception as e:
                sig = cls.__signature__
                bound_args = sig.bind(*args, **kwargs)
                logger.error(
                    "Initialization failed for %s instance with args %s",
                    cls,
                    bound_args,
                    exc_info=self.__log_tracebacks__,
                )
        else:
            init(self, *args, **kwargs)

    return __init__



# coding:utf-8
import json
import configparser
from warnings import warn
from collections.abc import Mapping
from .exceptions import ConfigNotSerializable

NoneType = type(None)

# inferred-type .ini load/dump methods

# section in which to dump top-level values that aren't mappings
INI_DEFAULT_SECTION = "__TOP_LEVEL__"


class INIConfigUnserializable(ConfigNotSerializable):
    pass


def load_ini(file):
    parser = configparser.ConfigParser()
    parser.read_file(file)
    keys = parser.sections()
    conf = {
        s: {k: _dynamic_str_parse(parser.get(s, k)) for k in parser.options(s)}
        for s in keys
    }
    top_level = conf.pop(INI_DEFAULT_SECTION, None)
    if top_level:
        update = set(top_level.keys()).difference(conf)
        conf.update([(k, top_level[k]) for k in update])
        if len(update) < len(top_level):
            warn(
                "The keys {} were defined at the top level in {} and also as subsections; they will be overridden by "
                "the subsections".format(tuple(set(top_level).difference(update)), file)
            )
    return conf


def dump_ini(conf, file):
    def is_str_mapping(d):
        return isinstance(d, Mapping) and all(isinstance(key, str) for key in d.keys())

    if not is_str_mapping(conf) or any(
        (not isinstance(v, (str, int, bool, NoneType, float, list)))
        and (not is_str_mapping(v))
        for v in conf.values()
    ):
        raise INIConfigUnserializable(
            "Can only dump Mapping[str, Union[Mapping[str, Any], str, int, float, bool, NoneType, list]] "
            "in .ini format; got {}".format(conf)
        )

    dumper = configparser.ConfigParser(default_section=INI_DEFAULT_SECTION)

    for name, section in conf.items():
        if isinstance(section, Mapping):
            dumper.add_section(name)
            for k, v in section.items():
                dumper.set(name, k, _dynamic_repr(v))
        else:
            # no section
            dumper.set(None, name, _dynamic_repr(section))

    dumper.write(file)


def _dynamic_str_parse(s):
    for parser in int, float, _parse_bool, _parse_none, json.loads, str:
        try:
            val = parser(s)
        except (TypeError, ValueError, KeyError):
            continue
        else:
            break
    else:
        raise TypeError("Could not parse {} dynamically".format(s))
    return val


def _parse_bool(s):
    return {"true": True, "false": False, "True": True, "False": False}[s]


def _parse_none(s):
    return {"none": None, "None": None, "null": None, "Null": None}[s]


def _dynamic_repr(o):
    if isinstance(o, (str, list, NoneType, bool)):
        return json.dumps(o)
    else:
        return repr(o)



# coding:utf-8
from inspect import signature
from bourbaki.introspection.types import eval_type_tree, concretize_typevars


class TypedIOTypeError(TypeError):
    def __init__(self, type_, *args):
        super().__init__(type_, *args)
        self.type_ = concretize_typevars(eval_type_tree(type_))


class TypedIOValueError(ValueError):
    def __init__(self, type_, value, exc=None):
        super().__init__(type_, value)
        self.type_ = concretize_typevars(eval_type_tree(type_))
        self.value = value
        self.exc = exc


class ExceptionReprMixin:
    def __repr__(self):
        return "{}({})".format(type(self).__name__, repr(str(self)))


class IOUndefinedForType(TypedIOTypeError, ExceptionReprMixin):
    source = None
    msg = "{}I/O is not defined for type {}"

    def __str__(self):
        source = "{} ".format(self.source.title()) if self.source else ""
        msg = self.msg.format(source, self.type_)
        if self.args:
            msg = msg + "; {}".format(self.args)
        return msg


class TypedInputError(TypedIOValueError, ExceptionReprMixin):
    source = None

    def __str__(self):
        source = "{} ".format(self.source) if self.source else ""
        msg = "Cannot parse type {} from {}value {}".format(
            self.type_, source, repr(self.value)
        )
        if self.exc is None:
            return msg
        return msg + "; raised {}".format(self.exc)


class TypedOutputError(TypedIOValueError, ExceptionReprMixin):
    source = None

    def __str__(self):
        source = "{} ".format(self.source) if self.source else ""
        msg = "Cannot write value {} to {}with target type {}".format(
            repr(self.value), source, self.type_
        )
        if self.exc is None:
            return msg
        return msg + "; raised {}".format(self.exc)


class ConfigIOUndefined(IOUndefinedForType):
    source = "configuration"


class ConfigIOUndefinedForKeyType(ConfigIOUndefined):
    msg = (
        "{}I/O is undefined for Mapping types with keys of type {}; "
        "use config_key_encoder.register to define a custom method"
    )


class ConfigCollectionKeysNotAllowed(ConfigIOUndefinedForKeyType):
    msg = "{}I/O is undefined for Mapping types with Collection-typed keys: {}"


class CLIIOUndefined(IOUndefinedForType):
    source = "command line"


class CLINestedCollectionsNotAllowed(CLIIOUndefined):
    msg = "{}I/O is undefined for nested collection types: {}"


class EnvIOUndefined(IOUndefinedForType):
    source = "configuration"


class ConfigTypedInputError(TypedInputError):
    source = "configuration"


class ConfigUnionInputError(ConfigTypedInputError):
    pass


class ConfigCallableInputError(ConfigTypedInputError):
    def __str__(self):
        msg = super().__str__()
        try:
            sig = signature(self.value)
        except ValueError:
            return msg
        else:
            return msg + "; signature is {}".format(sig)


class ConfigTypedOutputError(TypedOutputError):
    source = "configuration"


class ConfigTypedKeyOutputError(ConfigTypedOutputError):
    source = "configuration mapping key"


class ConfigUnionOutputError(ConfigTypedOutputError):
    pass


class CLITypedInputError(TypedInputError):
    source = "command line"


class CLIUnionInputError(CLITypedInputError):
    pass


class EnvTypedInputError(TypedInputError):
    source = "environment variable"



# coding:utf-8
import re
from ..reutils import (
    pathname_re,
    filename_re,
    any_name_re,
    upper_name_re,
    py_name_dynamic_re,
    py_dot_name_re,
    int_re,
    float_re,
)


stacktrace_start_re = re.compile(r"^\s*Traceback \(most recent call last\):")

# logging module standard field name -> (printf format type, regex, postprocessor)
log_fields = dict(
    asctime="s",
    created="f",
    filename="s",
    funcName="s",
    levelname="s",
    levelno="d",
    lineno="d",
    module="s",
    msecs="d",
    message="s",
    name="s",
    pathname="s",
    process="d",
    processName="s",
    relativeCreated="d",
    thread="d",
    threadName="s",
)

log_fmt_field_pat = r"(?<!\\)%\(({})\)({})".format(
    "|".join(log_fields), "|".join(set(log_fields.values()))
)

string_field_regexes = dict(
    filename=filename_re,
    funcName=py_name_dynamic_re,
    levelname=upper_name_re,
    module=py_dot_name_re,
    message=r".*",
    name=any_name_re,
    pathname=pathname_re,
    processName=any_name_re,
    threadName=any_name_re,
)

field_regexes = dict(f=float_re, d=int_re)

field_converters = dict(s=str, d=int, f=float)



# coding:utf-8
from shlex import shlex
from .cli_parse import cli_parser
from .cli_nargs_ import cli_nargs


def unquote(s):
    for qchar in ('"', "'", '"""', "'''"):
        if s.startswith(qchar) and s.endswith(qchar):
            return s[1:-1]
    return s


def lex_env_var(s):
    return list(map(unquote, shlex(s)))


class env_parser:
    def __init__(self, type_):
        self.parser = cli_parser(type_)
        self.nargs = cli_nargs(type_)

    def __call__(self, env_var):
        if self.nargs is None:
            return self.parser(env_var)

        return self.parser(lex_env_var(env_var))



# coding:utf-8
import decimal
import enum
import fractions
import io
import os
import sys
import typing
import types
import numbers
import pathlib
import datetime
import ipaddress
import collections
import uuid
from functools import partial
from operator import attrgetter
from urllib.parse import ParseResult as URL, urlunparse
from bourbaki.introspection.callables import function_classpath
from bourbaki.introspection.types import LazyType, NamedTupleABC
from bourbaki.introspection.generic_dispatch import (
    GenericTypeLevelSingleDispatch,
    UnknownSignature,
)
from bourbaki.introspection.generic_dispatch_helpers import (
    CollectionWrapper,
    TupleWrapper,
    MappingWrapper,
    UnionWrapper,
    LazyWrapper,
)
from .parsers import EnumParser, FlagParser
from .exceptions import (
    ConfigIOUndefined,
    ConfigTypedOutputError,
    ConfigTypedKeyOutputError,
    ConfigUnionOutputError,
)
from .utils import (
    Empty,
    identity,
    File,
    IODispatch,
    TypeCheckOutput,
    TypeCheckOutputFunc,
    TypeCheckOutputType,
)


class ConfigEncodeDispatch(IODispatch):
    exc_type = ConfigTypedOutputError


class ConfigKeyEncodeDispatch(IODispatch):
    exc_type = ConfigTypedKeyOutputError


to_int_config = ConfigEncodeDispatch("to_int_config", int)
to_int_config.register(numbers.Integral)(int)

to_bool_config = ConfigEncodeDispatch("to_bool_config", bool)
to_bool_config.register(numbers.Integral)(bool)

to_float_config = ConfigEncodeDispatch("to_float_config", float)
to_float_config.register(numbers.Number)(float)

to_complex_config = ConfigEncodeDispatch("to_complex_config", complex)
to_complex_config.register(numbers.Complex)(str)


@to_complex_config.register(numbers.Number)
def number_to_complex_str(x):
    return str(complex(x))


to_fraction_config = ConfigEncodeDispatch("to_fraction_config", fractions.Fraction)
to_fraction_config.register(fractions.Fraction)(str)


@to_fraction_config.register(numbers.Number)
def number_to_fraction_str(x):
    return str(fractions.Fraction(x))


to_decimal_config = ConfigEncodeDispatch("to_decimal_config", decimal.Decimal)
to_decimal_config.register(decimal.Decimal)(str)


@to_decimal_config.register(numbers.Number)
def number_to_decimal_str(x):
    return str(decimal.Decimal(x))


to_bytes_config = ConfigEncodeDispatch("to_bytes_config", bytes)
to_bytes_config.register((bytes, bytearray))(list)


bool_to_str = ConfigKeyEncodeDispatch("bool_to_str", bool)
bool_to_str.register(bool)(str)


@bool_to_str.register(numbers.Integral)
def integral_to_bool_str(i):
    return str(bool(i))


int_to_str = ConfigKeyEncodeDispatch("int_to_str", int)
int_to_str.register(int)(str)


@int_to_str.register(numbers.Integral)
def integral_to_int_str(b):
    return str(int(b))


float_to_str = ConfigKeyEncodeDispatch("float_to_str", float)
float_to_str.register(float)(str)


@float_to_str.register(numbers.Number)
def number_to_float_str(i):
    return str(float(i))


bytes_to_str = ConfigKeyEncodeDispatch("bytes_to_str", typing.ByteString)
bytes_to_str.register(bytes)(repr)


@bytes_to_str.register(bytearray)
def bytearray_to_str(b):
    return repr(bytes(b))


def repr_range(r):
    if r.step == 1:
        return "{}:{}".format(r.start, r.stop)
    return "{}:{}:{}".format(r.start, r.stop, r.step)


def encode_bytes_re(r):
    s = encode_str_re(r)
    if isinstance(s, str):
        return s
    return s.decode()


encode_str_re = attrgetter("pattern")


def config_file_encode(file: io.IOBase):
    if file in (sys.stdout, sys.stderr, sys.stdin):
        return "-"
    try:
        name = file.name
    except AttributeError as e:
        raise ConfigTypedOutputError(type(file), file, e)
    else:
        return os.path.abspath(name)


config_encoder_methods = {
    str: identity,
    range: repr_range,
    pathlib.Path: str,
    uuid.UUID: str,
    File: config_file_encode,
    types.FunctionType: function_classpath,
    types.BuiltinFunctionType: function_classpath,
    datetime.date: datetime.date.isoformat,
    datetime.datetime: datetime.datetime.isoformat,
    ipaddress.IPv4Address: str,
    ipaddress.IPv6Address: str,
    URL: urlunparse,
    typing.Pattern: encode_str_re,
    typing.Pattern[str]: encode_str_re,
    typing.Pattern[bytes]: encode_bytes_re,
}


class TypeCheckConfigEncoder(TypeCheckOutput):
    exc_cls = ConfigTypedOutputError


# don't isolate user-defined Generics because in general we don't know how to encode them...

config_key_encoder = GenericTypeLevelSingleDispatch(
    __name__, isolated_bases=[typing.Union]
)

config_encoder = GenericTypeLevelSingleDispatch(__name__, isolated_bases=[typing.Union])


@config_encoder.register_all(types.FunctionType, types.BuiltinFunctionType)
@config_key_encoder.register_all(types.FunctionType, types.BuiltinFunctionType)
class TypeCheckConfigEncodeFunc(TypeCheckOutputFunc):
    exc_cls = ConfigTypedOutputError


@config_encoder.register(typing.Type)
@config_key_encoder.register(typing.Type)
class TypeCheckConfigEncodeType(TypeCheckOutputType):
    exc_cls = ConfigTypedOutputError


# base for decoders that decode collections
class GenericConfigEncoderMixin:
    getter = config_encoder
    legal_container_types = None
    reduce = list
    helper_cls = None

    def typecheck(self, value):
        if self.legal_container_types is None:
            return value
        if not isinstance(value, self.legal_container_types):
            raise ConfigTypedOutputError(self.reduce, value)
        return value

    def __call__(self, value):
        arg = self.typecheck(value)
        return self.helper_cls.__call__(self, arg)


@config_encoder.register(typing.Collection)
class ConfigCollectionEncoder(GenericConfigEncoderMixin, CollectionWrapper):
    reduce = list
    helper_cls = CollectionWrapper


@config_encoder.register(typing.Mapping)
class ConfigMappingEncoder(GenericConfigEncoderMixin, MappingWrapper):
    reduce = dict
    key_getter = config_key_encoder
    helper_cls = MappingWrapper


@config_encoder.register(typing.ChainMap)
class ConfigChainMapEncoder(ConfigMappingEncoder):
    def __call__(self, arg):
        if isinstance(arg, collections.ChainMap):
            maps = arg.maps
        else:
            maps = [self.typecheck(arg)]
        to_map = super(type(self), self).__call__
        maps = map(to_map, maps)
        return list(maps)


@config_encoder.register(typing.Counter)
class ConfigCounterEncoder(ConfigMappingEncoder):
    def __init__(self, coll_type, key_type):
        super(ConfigCounterEncoder, self).__init__(coll_type, key_type, int)


@config_encoder.register(typing.Tuple)
class ConfigTupleEncoder(TupleWrapper):
    getter = config_encoder
    reduce = staticmethod(list)
    _collection_cls = ConfigCollectionEncoder


class _DictFromNamedTupleIter:
    def __init__(self, tuple_cls):
        self.tuple_cls = tuple_cls

    def __call__(self, iter_):
        return dict(zip(self.tuple_cls._fields, iter_))


@config_encoder.register(NamedTupleABC)
class ConfigNamedTupleEncoder(TupleWrapper):
    getter = config_encoder
    get_reducer = staticmethod(_DictFromNamedTupleIter)


@config_encoder.register(typing.Union)
class UnionConfigEncoder(GenericConfigEncoderMixin, UnionWrapper):
    tolerate_errors = (ConfigIOUndefined, UnknownSignature)
    reduce = staticmethod(next)
    helper_cls = UnionWrapper
    exc_class = ConfigUnionOutputError


@config_encoder.register(LazyType)
class LazyConfigDecoder(LazyWrapper):
    getter = config_encoder


@config_encoder.register_all(enum.Enum, enum.IntEnum)
@config_key_encoder.register_all(enum.Enum, enum.IntEnum)
def config_enum_encoder(enum_):
    return EnumParser(enum_).config_encode


@config_encoder.register_all(enum.Flag, enum.IntFlag)
@config_key_encoder.register_all(enum.Flag, enum.IntFlag)
def config_flag_encoder(enum_):
    return FlagParser(enum_).config_encode


@config_key_encoder.register(typing.Union)
class UnionConfigKeyEncoder(UnionConfigEncoder):
    getter = config_key_encoder


# typecheck the inputs of these; they aren't dispatched and they're specific
for t, enc in config_encoder_methods.items():
    f = partial(TypeCheckConfigEncoder, encoder=enc)
    config_encoder.register(t)(f)
    config_key_encoder.register(t)(f)


# don't need to typecheck these; they're dispatched selectively
for t, (enc, key_enc) in {
    int: (to_int_config, int_to_str),
    float: (to_float_config, float_to_str),
    bool: (to_bool_config, bool_to_str),
    typing.ByteString: (to_bytes_config, bytes_to_str),
}.items():
    config_encoder.register(t, as_const=True)(enc)
    config_key_encoder.register(t, as_const=True)(key_enc)


# let unspecified-type values pass through unaltered; dumping config later will raise an error if they aren't
# encodeable in the specified format. In most cases this will enforce JSON-typed objects
config_encoder.register(Empty, as_const=True)(identity)
config_encoder.register(typing.Any, as_const=True)(identity)


# require keys to be strings when the types aren't specified; deserialization can only return string keys so we require
# this for round-trip consistency
require_str_for_key = TypeCheckConfigEncoder(str, encoder=identity)
config_key_encoder.register(Empty, as_const=True)(require_str_for_key)
config_key_encoder.register(typing.Any, as_const=True)(require_str_for_key)
